"""Author: Jorrit Bakker.

Module handling data input in the form of a dictionary.
"""

import warnings
from dataclasses import dataclass
import numpy as np
from mibitrans.data.check_input import MissingValueError
from mibitrans.data.check_input import validate_input_values
from mibitrans.data.check_input import validate_source_zones
from mibitrans.visualize.show_conditions import source_zone


@dataclass
class HydrologicalParameters:
    """Dataclass handling input of hydrological parameters.

    Args:
        velocity (float) : Flow velocity in the direction of the groundwater gradient, in [m/d]. Optional if h_gradient
            and h_conductivity are specified.
        h_gradient (float) : Hydraulic gradient of the groundwater, in [m/m]. Optional if velocity is specified.
        h_conductivity (float) : Hydraulic conductivity of the aquifer, in [m/d]. Optional if velocity is specified.
        porosity (float) : Effective soil porosity [-]
        alpha_x (float) : The dispersivity in the x (longitudinal) direction in [m]
        alpha_y (float) : The dispersivity in the y (transverse-horizontal) direction in [m]
        alpha_z (float, optional) : The dispersivity in the z (transverse-vertical) direction in [m]. Defaults to 1e-10
        verbose (bool, optional): Verbose mode. Defaults to False.

    Raises:
        ValueError : If input parameters are incomplete or outside the valid domain.
        TypeError : If input parameters of incorrect datatype.
    """

    velocity: float = None
    h_gradient: float = None
    h_conductivity: float = None
    porosity: float = None
    alpha_x: float = None
    alpha_y: float = None
    alpha_z: float = 1e-10
    verbose: bool = False

    def __setattr__(self, parameter, value):
        """Override parent method to validate input when attribute is set."""
        validate_input_values(parameter, value)
        super().__setattr__(parameter, value)

    def __post_init__(self):
        """Check argument presence, types and domain. Calculate velocity if not given."""
        self._validate_input_presence()

        # Velocity is calculated from hydraulic gradient and conductivity when both are given.
        if self.h_gradient and self.h_conductivity:
            # Giving h_gradient & h_conductivity more specific than giving velocity. Input velocity will be overridden.
            if self.velocity is not None:
                warnings.warn(
                    "Both velocity and h_gradient & h_conductivity are defined. Value for velocity will be overridden.",
                    UserWarning,
                )
            self.velocity = self.h_gradient * self.h_conductivity / self.porosity
            if self.verbose:
                print(f"Groundwater flow velocity has been calculated to be {self.velocity} m/d.")

    def _validate_input_presence(self):
        missing_arguments = []
        if self.porosity is None:
            missing_arguments.append("porosity")
        if self.alpha_x is None:
            missing_arguments.append("alpha_x")
        if self.alpha_y is None:
            missing_arguments.append("alpha_y")

        if len(missing_arguments) > 0:
            raise MissingValueError(
                f"HydrologicalParameters missing {len(missing_arguments)} arguments: {missing_arguments}."
            )

        if self.velocity is None and (self.h_gradient is None or self.h_conductivity is None):
            raise MissingValueError(
                "HydrologicalParameters missing required arguments: either velocity or both h_gradient and"
                "h_conductivity."
            )

        if self.verbose:
            print("All required hydrological input arguments are present.")


@dataclass
class AttenuationParameters:
    """Dataclass handling parameters related to adsorption, diffusion and degradation.

    Args:
        retardation (float) : Retardation factor for transported contaminant [-]. Default is 1.
        decay_rate (float) : First order (linear) decay coefficient in [1/day]. Only required for linear decay models.
            Default is 0. Also sets corresponding half life.
        half_life (float) : Contaminant half life for 1st order (linear) decay, in [days]. Only required for
            linear decay models. Default is 0. Also sets corresponding decay_rate.
        diffusion (float) : Molecular diffusion [m2/day]. Default is 0.
        bulk_density (float) : Soil bulk density, in [g/m^3]. Optional if retardation is specified.
        partition_coefficient (float) : Partition coefficient of the transported contaminant to soil organic matter,
            in [m^3/g]. Optional if retardation is specified.
        fraction_organic_carbon (float) : Fraction of organic material in the soil [-].
            Optional if retardation is specified.
        verbose (bool, optional): Verbose mode. Defaults to False.

    Methods:
        calculate_retardation : Calculate retardation factor from bulk density, partition coefficient and
            fraction organic carbon when given porosity [-]

    Raises:
        ValueError : If input parameters are incomplete or outside the valid domain.
        TypeError : If input parameters of incorrect datatype.
    """

    retardation: float = 1
    decay_rate: float = 0
    half_life: float = 0
    diffusion: float = 0
    bulk_density: float = None
    partition_coefficient: float = None
    fraction_organic_carbon: float = None
    verbose: bool = False

    def __setattr__(self, parameter, value):
        """Override parent method to validate input when attribute is set."""
        validate_input_values(parameter, value)
        # Separate setattr for decay rate and half life because they should always be equivalent
        if parameter == "decay_rate" or parameter == "half_life":
            decay_rate, half_life = self._set_decay(parameter, value)
            super().__setattr__("decay_rate", decay_rate)
            super().__setattr__("half_life", half_life)
        else:
            super().__setattr__(parameter, value)

    def __post_init__(self):
        """Check argument presence, types and domain."""
        self.initialized = True

    def calculate_retardation(self, porosity: float):
        """Calculate retardation factor from soil adsorption parametrers and porosity."""
        self.retardation = (
            1 + (self.bulk_density / porosity) * self.partition_coefficient * self.fraction_organic_carbon
        )
        if self.verbose:
            print(f"Retardation factor has been calculated to be {self.retardation}.")

    def _require_linear_decay(self):
        if self.decay_rate is None and self.half_life is None:
            raise MissingValueError("Linear reaction model requires decay rate or half life.")

    def _set_decay(self, parameter, value):
        if parameter == "decay_rate" and (value != 0 or hasattr(self, "initialized")):
            decay_rate = value
            if value != 0:
                half_life = np.log(2) / value
            else:
                half_life = 0
        elif parameter == "half_life" and (value != 0 or hasattr(self, "initialized")):
            half_life = value
            if value != 0:
                decay_rate = np.log(2) / value
            else:
                decay_rate = 0
        elif parameter == "decay_rate":
            decay_rate = value
            half_life = 0
        elif parameter == "half_life":
            decay_rate = self.decay_rate
            half_life = self.half_life
        else:
            decay_rate = 0
            half_life = 0

        if self.decay_rate != decay_rate and self.decay_rate != 0 and not hasattr(self, "initialized") and value != 0:
            warnings.warn(
                "Both contaminant decay rate and half life were defined, but are not equal. "
                "Value for decay rate will be used.",
                UserWarning,
            )
            half_life = np.log(2) / self.decay_rate
            decay_rate = self.decay_rate

        return decay_rate, half_life


@dataclass
class SourceParameters:
    """Dataclass handling source parameters. Specifying concentrations and extent of source zone.

    Args:
        source_zone_boundary (np.ndarray) : Outer boundary of each source zone, in transverse horizontal direction
            (y-coordiante) [m]. y=0 is at the middle of the contaminant source. Input as numpy array of length equal
            to the amount of source zone. Last value in the array is the limit of the source. For a source with a single
            source zone, only one value is required. Source is symmetrical in the x-axis.
        source_zone_concentration (np.ndarray) : Contaminant concentration in each source zone [g/m^3]. Input as numpy
            array in the same order and of the same length as specified in source_zone_boundary.
        depth (float) : Depth (transverse vertical or z-dimension) of the source zone in [m].
        total_mass (float | str) : Mass of contaminant present in source zone, either expressed in [g],
            or set to 'infinite'. The latter meaning that the source mass and therefore, the source zone concentrations
            do not diminish over time.
        verbose (bool, optional): Verbose mode. Defaults to False.

    Raises:
        ValueError : If input parameters are incomplete or outside the valid domain.
        TypeError : If input parameters of incorrect datatype.
    """

    source_zone_boundary: np.ndarray = None
    source_zone_concentration: np.ndarray = None
    depth: float = None
    total_mass: float | str = "infinite"
    verbose: bool = False

    def __setattr__(self, parameter, value):
        """Override parent method to validate input when attribute is set."""
        validate_input_values(parameter, value)
        if parameter == "total_mass" and (isinstance(value, str) or value == np.inf):
            value = np.inf
        super().__setattr__(parameter, value)
        # When setting source zone boundary or concentration, and both present, check validity in respect to each other.
        if parameter in ["source_zone_boundary", "source_zone_concentration"] and (
            self.source_zone_boundary is not None and self.source_zone_concentration is not None
        ):
            boundary, concentration = validate_source_zones(self.source_zone_boundary, self.source_zone_concentration)
            super().__setattr__("source_zone_boundary", boundary)
            super().__setattr__("source_zone_concentration", concentration)

    def __post_init__(self):
        """Check argument presence, types and domain."""
        self._validate_input_presence()

        # Make sure naming for infinite source mass is consistent from this point onward
        if isinstance(self.total_mass, str):
            self.total_mass = "infinite"

    def interpolate(self, n_zones, method):
        """Rediscretize source to n zones. Either through linear interpolation or using a normal distribution."""
        warnings.warn("This functionality is not implemented yet. Try again later.")
        return None

    def visualize(self):
        """Plot the source zone concentration distribution."""
        source_zone(self)

    def _validate_input_presence(self):
        # Check if all required arguments are present
        missing_arguments = []
        if self.source_zone_boundary is None:
            missing_arguments.append("source_zone_boundary")
        if self.source_zone_concentration is None:
            missing_arguments.append("source_zone_concentration")
        if self.depth is None:
            missing_arguments.append("depth")

        if len(missing_arguments) > 0:
            raise MissingValueError(
                f"SourceParameters missing {len(missing_arguments)} arguments: {missing_arguments}."
            )


@dataclass
class ModelParameters:
    """Dataclass handling model discretization parameters.

    Args:
        model_length (float) : Model extent in the longitudinal (x) direction in [m].
        model_width (float) : Model extent in the transverse horizontal (y) direction in [m].
        model_time (float) : Model duration in [days].
        dx (float, optional) : Model grid discretization step size in the longitudinal (x) direction, in [m]. By
            default, dx = (model_length / 100).
        dy (float, optional) : Model grid discretization step size in the transverse horizontal (y) direction, in [m].
            By default, dy = (model_width / 50).
        dt (float, optional) : Model time discretization step size, in [days]. By default, dt = (model_time / 10).
        verbose (bool, optional): Verbose mode. Defaults to False.

    Raises:
        ValueError : If input parameters are incomplete or outside the valid domain.
        ValueError : If model dimensions are smaller than their given step size.
        TypeError : If input parameters of incorrect datatype.

    """

    model_length: float
    model_width: float
    model_time: float
    dx: float = None
    dy: float = None
    dt: float = None
    verbose: bool = False

    def __setattr__(self, parameter, value):
        """Override parent method to validate input when attribute is set."""
        validate_input_values(parameter, value)
        super().__setattr__(parameter, value)
        self._validate_stepsize(parameter)

    def _validate_stepsize(self, parameter):
        """Validate if model step size is not larger than the corresponding model dimension."""
        match parameter:
            case "dx" | "model_length":
                if self.dx is not None and self.model_length is not None:
                    if self.dx > self.model_length:
                        raise ValueError(
                            f"Model x-direction step size ({self.dx}) "
                            f"is greater than the model length ({self.model_length})."
                        )
            case "dy" | "model_width":
                if self.dy is not None and self.model_width is not None:
                    if self.dy > self.model_width:
                        raise ValueError(
                            f"Model y-direction step size ({self.dy}) "
                            f"is greater than the model width ({self.model_width})."
                        )
            case "dt" | "model_time":
                if self.dt is not None and self.model_time is not None:
                    if self.dt > self.model_time:
                        raise ValueError(
                            f"Model time step size ({self.dt}) "
                            f"is greater than the total model time ({self.model_time})."
                        )

    def __post_init__(self):
        """Set model discretization parameters if not provided."""
        if not self.dx:
            self.dx = self.model_length / 100
        if not self.dy:
            self.dy = self.model_width / 50
        if not self.dt:
            self.dt = self.model_time / 10



"""Documentation about mibitrans."""

# Add some commonly used functions as top-level imports
from mibitrans.data.parameters import HydrologicalParameters
from mibitrans.data.parameters import AttenuationParameters
from mibitrans.data.parameters import SourceParameters
from mibitrans.data.parameters import ModelParameters
from mibitrans.data.parameter_information import ElectronAcceptors
from mibitrans.data.parameter_information import UtilizationFactor
from mibitrans.transport.models import Anatrans
from mibitrans.transport.models import Bioscreen
from mibitrans.transport.models import Mibitrans
from mibitrans.visualize.plot_line import centerline
from mibitrans.visualize.plot_line import transverse
from mibitrans.visualize.plot_line import breakthrough
from mibitrans.visualize.plot_surface import plume_2d
from mibitrans.visualize.plot_surface import plume_3d

import logging

logging.getLogger(__name__).addHandler(logging.NullHandler())

__author__ = "Alraune Zech"
__email__ = "a.zech@uu.nl"
__version__ = "0.5.0"



import warnings
import numpy as np
from scipy.integrate import quad
from scipy.integrate import quad_vec
from scipy.special import erf
from scipy.special import erfc
from scipy.special import erfcx
from mibitrans.data.check_input import validate_input_values
from mibitrans.transport.model_parent import Results
from mibitrans.transport.model_parent import Transport3D


class Mibitrans(Transport3D):
    """Model class using an exact analytical solution as described in Karanovic (2007), based on Wexler (1992).

    Karanovic (2007) implemented the Wexler (1992) exact analytical solution in the Excel based BIOSCREEN-AT, and added
    source depletion, akin to that implemented in its predecessor BIOSCREEN by Newell et al. (1997). The Mibitrans model
    allows for the same method as used in BIOSCREEN-AT, but expands it by allowing multiple source zones (by means of
    superposition) and including the instant reaction model. These were present in the original BIOSCREEN, but not
    reimplemented in BIOSCREEN-AT. Using a single source zone in this model, and not using the instant reaction option
    will make the Mibitrans solution resolve to the equation described in Karanovic (2007). Which in turn resolves to
    the Wexler (1992) solution if source depletion is disabled.

    Karanovic, M., Neville, C. J., & Andrews, C. B. (2007). BIOSCREEN‐AT: BIOSCREEN with an exact analytical solution.
    Groundwater, 45(2), 242-245.

    Newell, C. J., McLeod, R. K., & Gonzales, J. R. (1997). BIOSCREEN natural attenuation decision support
    system version 1.4 revisions, Tech. rep., U.S. EPA.

    Wexler, E. J. (1992). Analytical solutions for one-, two-, and three-dimensional solute transport in ground-water
    systems with uniform flow. US Government Printing Office.
    """

    def __init__(
        self,
        hydrological_parameters,
        attenuation_parameters,
        source_parameters,
        model_parameters,
        verbose=False,
    ):
        """Initialize model object.

        Args:
            hydrological_parameters (mibitrans.data.parameters.HydrologicalParameters) : Dataclass object containing
                hydrological parameters from HydrologicalParameters.
            attenuation_parameters (mibitrans.data.read.AttenuationParameters) : Dataclass object containing adsorption,
                degradation and diffusion parameters from AttenuationParameters.
            source_parameters (mibitrans.data.read.SourceParameters) : Dataclass object containing source parameters
                from SourceParameters.
            model_parameters (mibitrans.data.read.ModelParameters) : Dataclass object containing model parameters from
                ModelParameters.
            verbose (bool, optional): Verbose mode. Defaults to False.

        Attributes:
            mode (str) : Current model mode. Is 'linear' by default. Once instant reaction parameters are provided. Use
                this attribute to switch between 'linear' and 'instant_reaction' models.
            cxyt (np.ndarray) : Output array containing concentrations in model domain, in [g/m^3]. Indexed as [t,y,x]
            relative_cxyt (np.ndarray) : Output array with concentrations in model domain, divided by the maximum source
                zone concentration at t=0. Indexed as [t,y,x].
            x (np.ndarray) : Discretized model x-dimension, in [m].
            y (np.ndarray) : Discretized model y-dimension, in [y].
            t (np.ndarray) : Discretized model t-dimension, in [days].
            c_source (np.ndarray) : Nett source zone concentrations, accounting for source superposition, in [g/m^3].
            vr (float) : Retarded groundwater flow velocity, in [m/d].
            k_source (float) : Source zone decay rate, in [1/days].
            biodegradation_capacity (float) : Maximum capacity of biodegradation given provided electron acceptor
            concentrations, in [g/m^3].

        Methods:
            run : Run model with current parameters, returns Results object.
            sample : Calculate concentration at any given position and point in time.
            instant_reaction : Activate the instant reaction model by providing electron acceptor concentrations. And
                optionally electron acceptor utilization factors. Switch between model modes by using the mode
                attribute.

        Raises:
            TypeError : If input is not of the correct Dataclass.

        """
        super().__init__(hydrological_parameters, attenuation_parameters, source_parameters, model_parameters, verbose)

    @property
    def short_description(self):
        """Return short description of model type."""
        if self.biodegradation_capacity:
            return "Mibitrans Instant Reaction"
        else:
            return "Mibitrans Linear"

    def run(self):
        """Calculate the concentration for all discretized x, y and t using the analytical transport model."""
        self._check_model_mode_before_run()
        with np.errstate(divide="ignore", invalid="ignore"):
            self.cxyt = self._calculate_concentration_for_all_xyt()
        return Results(self)

    def sample(self, x_position, y_position, time):
        """Give concentration at any given position and point in time.

        Args:
            x_position (float): x position in domain extent [m].
            y_position (float): y position in domain extent [m].
            time (float): time for which concentration is sampled [days].

        Returns:
            concentration (float): concentration at given position and point in time [g/m^3].

        """
        # Different sample method than parent class, as field-wide calculations use array indices
        for par, value in locals().items():
            if par != "self":
                validate_input_values(par, value)

        self._pre_run_initialization_parameters()

        def integrand(t, sz):
            div_term = 2 * np.sqrt(self.disp_y * t**4)
            inner_term = self._src_pars.depth / (2 * np.sqrt(self.disp_z * t**4))
            integrand_results = (
                1
                / (t**3)
                * (
                    np.exp(
                        (-self.k_source - self._decay_rate) * t**4
                        - (x_position - self.rv * t**4) ** 2 / (4 * self.disp_x * t**4)
                    )
                    * (
                        erfc((y_position - self.y_source[sz]) / div_term)
                        - erfc((y_position + self.y_source[sz]) / div_term)
                    )
                    * (erfc(-inner_term) - erfc(inner_term))
                )
            )
            return integrand_results

        conc_array = np.zeros(len(self.c_source))
        error_array = np.zeros(len(self.c_source))
        time = time ** (1 / 4)
        with np.errstate(divide="ignore", invalid="ignore"):
            for sz in range(len(self.c_source)):
                integral_term, error = quad(integrand, 0, time, limit=10000, args=(sz,))
                source_term = (
                    self.c_source[sz] * x_position / (8 * np.sqrt(np.pi * self.disp_x)) * np.exp(-self.k_source * time)
                )
                conc_array[sz] = 4 * integral_term * source_term
                error_array[sz] = error
            concentration = np.sum(conc_array)
            if self._mode == "instant_reaction":
                concentration -= self.biodegradation_capacity
                if concentration < 0:
                    concentration = 0
        return concentration

    def _pre_run_initialization_parameters(self):
        super()._pre_run_initialization_parameters()
        self.disp_x = self._hyd_pars.alpha_x * self.rv + self._att_pars.diffusion
        self.disp_y = self._hyd_pars.alpha_y * self.rv + self._att_pars.diffusion
        self.disp_z = self._hyd_pars.alpha_z * self.rv + self._att_pars.diffusion
        # self.integral_term = np.zeros(self.ttt.shape)
        # Stores integral error for each time step and source zone
        self.error_size = np.zeros((len(self._src_pars.source_zone_boundary), len(self.t)))

    def _calculate_concentration_for_all_xyt(self):
        cxyt = self.cxyt.copy()
        for sz in range(len(self.c_source)):
            integral_sum = self._equation_term_integral(sz)
            source_term = self._equation_term_source(sz)
            cxyt[:, :, 1:] += integral_sum[:, :, 1:] * source_term
            # If x=0, equation resolves to c=0, therefore, x=0 needs to be evaluated separately
            cxyt[:, :, 0] += self._equation_term_source_x_is_zero(sz)[:, :, 0]
        if self._mode == "instant_reaction":
            self.cxyt_noBC = cxyt.copy()
            cxyt -= self.biodegradation_capacity
            cxyt = np.where(cxyt < 0, 0, cxyt)
        return cxyt

    def _equation_term_integral(self, sz):
        integral_term = np.zeros(self.cxyt.shape)
        for j in range(len(self.t)):
            if self.verbose:
                print("integrating for source zone ", sz, " and t =", self.t[j], "days")
            if j == 0:
                lower_bound = 0
            else:
                lower_bound = self.t[j - 1]
            upper_bound = self.t[j]
            integral_term[j, :, 1:], self.error_size[sz, j] = quad_vec(
                self._equation_integrand, lower_bound, upper_bound, limit=10000 // len(self.t), args=(sz,)
            )
        integral_sum = np.cumsum(integral_term, axis=0)
        return integral_sum

    def _equation_integrand(self, t, sz):
        term = 1 / (t ** (3 / 2)) * self._equation_term_x(t) * self._equation_term_y(t, sz) * self._equation_term_z(t)
        term[np.isnan(term)] = 0
        return term

    def _equation_term_x(self, t):
        term = np.exp(
            (-self.k_source - self._decay_rate) * t - (self.xxx[:, :, 1:] - self.rv * t) ** 2 / (4 * self.disp_x * t)
        )
        term[np.isnan(term)] = 0
        return term

    def _equation_term_y(self, t, sz):
        div_term = 2 * np.sqrt(self.disp_y * t)
        term = erfc((self.yyy - self.y_source[sz]) / div_term) - erfc((self.yyy + self.y_source[sz]) / div_term)
        term[np.isnan(term)] = 0
        return term

    def _equation_term_z(self, t):
        if t == 0 or self.disp_z == 0:
            inner_term = 2
        else:
            inner_term = self._src_pars.depth / (2 * np.sqrt(self.disp_z * t))
        return erfc(-inner_term) - erfc(inner_term)

    def _equation_term_source(self, sz):
        return (
            self.c_source[sz]
            * self.xxx[:, :, 1:]
            / (8 * np.sqrt(np.pi * self.disp_x))
            * np.exp(-self.k_source * self.ttt)
        )

    def _equation_term_source_x_is_zero(self, sz):
        # Select y-positions of current source zone
        zone_location = np.where(abs(self.yyy) <= self.y_source[sz], 1, 0)
        return self.c_source[sz] * zone_location * np.exp(-self.k_source * self.ttt)


class Anatrans(Transport3D):
    """Model class using an analytical solution based on Bear (1979), Domenico (1987) & Newell et al. (1997).

    Under the assumption that C(x,y,z,t) = C(x,t) * C(y,t) * C(z,t), the 3D ADE can be broken up in three separate
    differential equations which can be solved individually. For C(x,t) the solution is given in Bear (1979), C(y,t) and
    C(z,t) can be derived from Crank (1975). The equation used for Anatrans is the combination of these solutions, with
    addition of source depletion, source superposition and instant reaction model, described in Newell et al. (1997) and
    implemented in the BIOSCREEN screening model. The solution of Newell et al. (1997) is based of the Domenico (1987)
    solution, a truncated version of the equation described above, which introduces an error with a size dependent on
    the ratio of flow velocity and longitudinal dispersivity. Anatrans instead uses the fully untruncated version.

    Bear, J. 1979. Hydraulics of Ground Water. New York: McGraw-Hill.

    Crank, J. 1975. The mathematics of Diffusion. New York: Oxford University Press.

    Domenico, P. A. (1987). An analytical model for multidimensional transport of a decaying contaminant species.
    Journal of Hydrology, 91(1-2), 49-58.

    Newell, C. J., McLeod, R. K., & Gonzales, J. R. (1997). BIOSCREEN natural attenuation decision support
    system version 1.4 revisions, Tech. rep., U.S. EPA.
    """

    def __init__(
        self,
        hydrological_parameters,
        attenuation_parameters,
        source_parameters,
        model_parameters,
        verbose=False,
    ):
        """Initialize model object.

        Args:
            hydrological_parameters (mibitrans.data.parameters.HydrologicalParameters) : Dataclass object containing
                hydrological parameters from HydrologicalParameters.
            attenuation_parameters (mibitrans.data.read.AttenuationParameters) : Dataclass object containing adsorption,
                degradation and diffusion parameters from AttenuationParameters.
            source_parameters (mibitrans.data.read.SourceParameters) : Dataclass object containing source parameters
                from SourceParameters.
            model_parameters (mibitrans.data.read.ModelParameters) : Dataclass object containing model parameters from
                ModelParameters.
            verbose (bool, optional): Verbose mode. Defaults to False.

        Attributes:
            mode (str) : Current model mode. Is 'linear' by default. Once instant reaction parameters are provided. Use
                this attribute to switch between 'linear' and 'instant_reaction' models.
            cxyt (np.ndarray) : Output array containing concentrations in model domain, in [g/m^3]. Indexed as [t,y,x]
            relative_cxyt (np.ndarray) : Output array with concentrations in model domain, divided by the maximum source
                zone concentration at t=0. Indexed as [t,y,x].
            x (np.ndarray) : Discretized model x-dimension, in [m].
            y (np.ndarray) : Discretized model y-dimension, in [y].
            t (np.ndarray) : Discretized model t-dimension, in [days].
            c_source (np.ndarray) : Nett source zone concentrations, accounting for source superposition, in [g/m^3].
            vr (float) : Retarded groundwater flow velocity, in [m/d].
            k_source (float) : Source zone decay rate, in [1/days].
            biodegradation_capacity (float) : Maximum capacity of biodegradation given provided electron acceptor
            concentrations, in [g/m^3].

        Methods:
            run : Run model with current parameters, returns Results object.
            sample : Calculate concentration at any given position and point in time.
            instant_reaction : Activate the instant reaction model by providing electron acceptor concentrations. And
                optionally electron acceptor utilization factors. Switch between model modes by using the mode
                attribute.
            centerline : Plot center of contaminant plume of this model, at a specified time and y position.
            transverse : Plot concentration distribution as a line horizontal transverse to the plume extent.
            breakthrough : Plot contaminant breakthrough curve at given x and y position in model domain.
            plume_2d : Plot contaminant plume as a 2D colormesh, at a specified time.
            plume_3d : Plot contaminant plume as a 3D surface, at a specified time.

        Raises:
            TypeError : If input is not of the correct Dataclass.

        """
        super().__init__(hydrological_parameters, attenuation_parameters, source_parameters, model_parameters, verbose)
        if self._att_pars.diffusion != 0:
            warnings.warn("Domenico model does not consider molecular diffusion.", UserWarning)

    @property
    def short_description(self):
        """Short description of model type."""
        return "Anatrans model"

    def run(self):
        """Calculate the concentration for all discretized x, y and t using the analytical transport model."""
        self._check_model_mode_before_run()
        with np.errstate(divide="ignore", invalid="ignore"):
            self.cxyt = self._calculate_concentration_for_all_xyt(self.xxx, self.yyy, self.ttt)
        return Results(self)

    def sample(self, x_position, y_position, time):
        """Give concentration at any given position and point in time.

        Args:
            x_position (float): x position in domain extent [m].
            y_position (float): y position in domain extent [m].
            time (float): time for which concentration is sampled [days].

        Returns:
            concentration (float): concentration at given position and point in time [g/m^3].

        """
        for par, value in locals().items():
            if par != "self":
                validate_input_values(par, value)

        self._pre_run_initialization_parameters()

        if self.mode == "instant_reaction":
            save_c_noBC = self.cxyt_noBC.copy()
        x = np.array([x_position])
        y = np.array([y_position])
        t = np.array([time])
        concentration = self._calculate_concentration_for_all_xyt(x, y, t)[0]
        if self.mode == "instant_reaction":
            self.cxyt_noBC = save_c_noBC
        return concentration

    def _equation_term_x(self, xxx, ttt, decay_sqrt):
        return np.exp(xxx * (1 - decay_sqrt) / (self._hyd_pars.alpha_x * 2)) * erfc(
            (xxx - self.rv * ttt * decay_sqrt) / (2 * np.sqrt(self._hyd_pars.alpha_x * self.rv * ttt))
        )

    def _equation_term_additional_x(self, xxx, ttt, decay_sqrt):
        erfc_inner = (xxx + decay_sqrt * self.rv * ttt) / (2 * np.sqrt(self._hyd_pars.alpha_x * self.rv * ttt))
        # Additional term is prone to overflow of exp and underflow of erfc under certain parameter combinations.
        # To decrease cases, used erfcx. Where erfcx(a) = exp(a**2)*erfc(a) -> exp(b)*erfc(a) = exp(b - a**2) * erfcx(a)
        term = np.exp(xxx * (1 + decay_sqrt) * (1 / 2) / self._hyd_pars.alpha_x - erfc_inner**2) * erfcx(erfc_inner)
        return term

    def _equation_term_z(self, xxx):
        inner_term = self._src_pars.depth / (2 * np.sqrt(self._hyd_pars.alpha_z * xxx))
        return erf(inner_term) - erf(-inner_term)

    def _equation_term_source_decay(self, xxx, ttt):
        term = np.exp(-self.k_source * (ttt - xxx / self.rv))
        # Term can be max 1; can not have 'generation' of solute ahead of advection.
        return np.where(term > 1, 1, term)

    def _equation_term_y(self, i, xxx, yyy):
        div_term = 2 * np.sqrt(self._hyd_pars.alpha_y * xxx)
        term = erf((yyy + self.y_source[i]) / div_term) - erf((yyy - self.y_source[i]) / div_term)
        term[np.isnan(term)] = 0
        return term

    def _calculate_concentration_for_all_xyt(self, xxx, yyy, ttt):
        cxyt = 0
        decay_sqrt = np.sqrt(1 + 4 * self._decay_rate * self._hyd_pars.alpha_x / self.rv)
        x_term = self._equation_term_x(xxx, ttt, decay_sqrt)
        additional_x = self._equation_term_additional_x(xxx, ttt, decay_sqrt)
        z_term = self._equation_term_z(xxx)
        source_decay = self._equation_term_source_decay(xxx, ttt)
        for i in range(len(self.c_source)):
            y_term = self._equation_term_y(i, xxx, yyy)
            cxyt_step = 1 / 8 * self.c_source[i] * source_decay * (x_term + additional_x) * y_term * z_term
            cxyt += cxyt_step
        if self._mode == "instant_reaction":
            self.cxyt_noBC = cxyt.copy()
            cxyt -= self.biodegradation_capacity
            cxyt = np.where(cxyt < 0, 0, cxyt)
        self.has_run = True
        return cxyt


# bioscreen ; Domenico - additional term


class Bioscreen(Anatrans):
    """Model class using the analytical solution implemented in the BIOSCREEN screening model, Newell et al. (1997).

    This model is an exact implementation of the transport equations implemented in the BIOSCREEN screening model of
    Newell et al. (1997), which is based on the Domenico (1987) analytical model. Using a truncated version of the
    equation used in the Anatrans model. This model is implemented as a method of comparison with the original BIOSCREEN
    software. And is included for legacy reasons, since it is the first model implemented in the mibitrans package,
    serving as a basis for the other models. However, caution should be taken when using this model, since a varying
    error is introduced by using the truncated analytical solution. The error is most prominent for shorter times and
    distances from the source, and depends on the ratio of flow velocity and longitudinal dispersivity. For modelling,
    the Anatrans (untruncated approximate solution) and Mibitrans (exact analytical solution) models are recommended
    instead.

    Domenico, P. A. (1987). An analytical model for multidimensional transport of a decaying contaminant species.
    Journal of Hydrology, 91(1-2), 49-58.

    Newell, C. J., McLeod, R. K., & Gonzales, J. R. (1997). BIOSCREEN natural attenuation decision support
    system version 1.4 revisions, Tech. rep., U.S. EPA.
    """

    def __init__(
        self,
        hydrological_parameters,
        attenuation_parameters,
        source_parameters,
        model_parameters,
        verbose=False,
    ):
        """Initialize model object.

        Args:
            hydrological_parameters (mibitrans.data.parameters.HydrologicalParameters) : Dataclass object containing
                hydrological parameters from HydrologicalParameters.
            attenuation_parameters (mibitrans.data.read.AttenuationParameters) : Dataclass object containing adsorption,
                degradation and diffusion parameters from AttenuationParameters.
            source_parameters (mibitrans.data.read.SourceParameters) : Dataclass object containing source parameters
                from SourceParameters.
            model_parameters (mibitrans.data.read.ModelParameters) : Dataclass object containing model parameters from
                ModelParameters.
            verbose (bool, optional): Verbose mode. Defaults to False.

        Attributes:
            mode (str) : Current model mode. Is 'linear' by default. Once instant reaction parameters are provided. Use
                this attribute to switch between 'linear' and 'instant_reaction' models.
            cxyt (np.ndarray) : Output array containing concentrations in model domain, in [g/m^3]. Indexed as [t,y,x]
            relative_cxyt (np.ndarray) : Output array with concentrations in model domain, divided by the maximum source
                zone concentration at t=0. Indexed as [t,y,x].
            x (np.ndarray) : Discretized model x-dimension, in [m].
            y (np.ndarray) : Discretized model y-dimension, in [y].
            t (np.ndarray) : Discretized model t-dimension, in [days].
            c_source (np.ndarray) : Nett source zone concentrations, accounting for source superposition, in [g/m^3].
            vr (float) : Retarded groundwater flow velocity, in [m/d].
            k_source (float) : Source zone decay rate, in [1/days].
            biodegradation_capacity (float) : Maximum capacity of biodegradation given provided electron acceptor
            concentrations, in [g/m^3].

        Methods:
            run : Run model with current parameters, returns Results object.
            sample : Calculate concentration at any given position and point in time.
            instant_reaction : Activate the instant reaction model by providing electron acceptor concentrations. And
                optionally electron acceptor utilization factors. Switch between model modes by using the mode
                attribute.

        Raises:
            TypeError : If input is not of the correct Dataclass.

        """
        super().__init__(hydrological_parameters, attenuation_parameters, source_parameters, model_parameters, verbose)

    @property
    def short_description(self):
        """Short description of model type."""
        return "Bioscreen model"

    def _calculate_concentration_for_all_xyt(self, xxx, yyy, ttt):
        # Difference with the Anatrans solution is the lack of additional term.
        cxyt = 0
        with np.errstate(divide="ignore", invalid="ignore"):
            decay_sqrt = np.sqrt(1 + 4 * self._decay_rate * self._hyd_pars.alpha_x / self.rv)
            x_term = self._equation_term_x(xxx, ttt, decay_sqrt)
            z_term = self._equation_term_z(xxx)
            source_decay = self._equation_term_source_decay(xxx, ttt)
            for i in range(len(self.c_source)):
                y_term = self._equation_term_y(i, xxx, yyy)
                cxyt_step = 1 / 8 * self.c_source[i] * source_decay * x_term * y_term * z_term
                cxyt += cxyt_step
        if self._mode == "instant_reaction":
            self.cxyt_noBC = cxyt.copy()
            cxyt -= self.biodegradation_capacity
            cxyt = np.where(cxyt < 0, 0, cxyt)
        self.has_run = True
        return cxyt



import copy
import numpy as np
import pytest
from mibitrans.data.parameters import AttenuationParameters
from mibitrans.data.parameters import HydrologicalParameters
from mibitrans.data.parameters import ModelParameters
from mibitrans.data.parameters import SourceParameters
from mibitrans.transport.models import Anatrans
from mibitrans.transport.models import Bioscreen
from mibitrans.transport.models import Mibitrans


@pytest.fixture(scope="session")
def test_hydro_pars():
    """HydrologicalParameters fixture with example data for tests."""
    return HydrologicalParameters(velocity=10 / 365, porosity=0.25, alpha_x=10, alpha_y=1, alpha_z=0.1)


@pytest.fixture(scope="session")
def test_att_pars():
    """AttenuationParameters fixture with example data for tests."""
    return AttenuationParameters(
        retardation=1,
        half_life=0.1 * 365,
    )


@pytest.fixture(scope="session")
def test_att_pars_nodecay():
    """AttenuationParameters fixture with example data for tests."""
    return AttenuationParameters(
        retardation=1,
        decay_rate=0,
    )


electron_acceptor_dict = dict(delta_oxygen=0.5, delta_nitrate=0.5, ferrous_iron=0.5, delta_sulfate=0.5, methane=0.5)


@pytest.fixture(scope="session")
def test_source_pars():
    """SourceParameters fixture with example data for tests."""
    return SourceParameters(
        source_zone_boundary=np.array([5, 10, 15]),
        source_zone_concentration=np.array([10, 5, 2]),
        depth=10,
        total_mass=1000000,
    )


@pytest.fixture(scope="session")
def test_model_pars():
    """ModelParameters fixture with example data for tests."""
    return ModelParameters(model_length=50, model_width=30, model_time=3 * 365, dx=10, dy=5, dt=1 * 365)


@pytest.fixture(scope="session")
def test_model_pars_short(test_source_pars, test_model_pars):
    """Model Parameters fixture with smaller model width for testing."""
    short_model_pars = copy.copy(test_model_pars)
    short_model_pars.model_width = test_source_pars.source_zone_boundary[-1] - 1
    return short_model_pars


@pytest.fixture(scope="session")
def test_mibitrans_model_nodecay(test_hydro_pars, test_att_pars_nodecay, test_source_pars, test_model_pars):
    """Mibitrans fixture model object for testing, with no decay."""
    obj = Mibitrans(test_hydro_pars, test_att_pars_nodecay, test_source_pars, test_model_pars)
    res = obj.run()
    return obj, res


@pytest.fixture(scope="session")
def test_mibitrans_model_lineardecay(test_hydro_pars, test_att_pars, test_source_pars, test_model_pars):
    """Mibitrans fixture model object for testing, with linear decay."""
    obj = Mibitrans(test_hydro_pars, test_att_pars, test_source_pars, test_model_pars)
    res = obj.run()
    return obj, res


@pytest.fixture(scope="session")
def test_mibitrans_model_instantreaction(test_hydro_pars, test_att_pars, test_source_pars, test_model_pars):
    """Mibitrans fixture model object for testing, with instant reaction."""
    obj = Mibitrans(test_hydro_pars, test_att_pars, test_source_pars, test_model_pars)
    obj.instant_reaction(electron_acceptors=electron_acceptor_dict)
    res = obj.run()
    return obj, res


@pytest.fixture(scope="module")
def test_anatrans_model_nodecay(test_hydro_pars, test_att_pars_nodecay, test_source_pars, test_model_pars):
    """Anatrans fixture model object for testing, with no decay."""
    obj = Anatrans(test_hydro_pars, test_att_pars_nodecay, test_source_pars, test_model_pars)
    res = obj.run()
    return obj, res


@pytest.fixture(scope="module")
def test_anatrans_model_lineardecay(test_hydro_pars, test_att_pars, test_source_pars, test_model_pars):
    """Anatrans fixture model object for testing, with linear decay."""
    obj = Anatrans(test_hydro_pars, test_att_pars, test_source_pars, test_model_pars)
    res = obj.run()
    return obj, res


@pytest.fixture(scope="module")
def test_anatrans_model_instantreaction(test_hydro_pars, test_att_pars, test_source_pars, test_model_pars):
    """Anatrans fixture model object for testing, with instant reaction."""
    obj = Anatrans(test_hydro_pars, test_att_pars, test_source_pars, test_model_pars)
    obj.instant_reaction(electron_acceptors=electron_acceptor_dict)
    res = obj.run()
    return obj, res


@pytest.fixture(scope="session")
def test_bioscreen_model_nodecay(test_hydro_pars, test_att_pars_nodecay, test_source_pars, test_model_pars):
    """Bioscreen fixture model object for testing, with no decay."""
    test_att_pars_nodecay.decay_rate = 0
    obj = Bioscreen(test_hydro_pars, test_att_pars_nodecay, test_source_pars, test_model_pars)
    res = obj.run()
    return obj, res


@pytest.fixture(scope="session")
def test_bioscreen_model_lineardecay(test_hydro_pars, test_att_pars, test_source_pars, test_model_pars):
    """Bioscreen fixture model object for testing, with linear decay."""
    obj = Bioscreen(test_hydro_pars, test_att_pars, test_source_pars, test_model_pars)
    res = obj.run()
    return obj, res


@pytest.fixture(scope="session")
def test_bioscreen_model_instantreaction(test_hydro_pars, test_att_pars, test_source_pars, test_model_pars):
    """Bioscreen fixture model object for testing, with instant reaction."""
    obj = Bioscreen(test_hydro_pars, test_att_pars, test_source_pars, test_model_pars)
    obj.instant_reaction(electron_acceptors=electron_acceptor_dict)
    res = obj.run()
    return obj, res



import matplotlib.pyplot as plt
import numpy as np
from matplotlib import animation
from mibitrans.data.check_input import check_model_type
from mibitrans.data.check_input import check_time_in_domain
from mibitrans.visualize.plot_line import _plot_title_generator
from mibitrans.visualize.plot_line import allowed_model_types

relative_conc_zlabel = r"Relative concentration ($C/C_0$)"
absolute_conc_zlabel = r"Concentration [g/$m^{3}$]"


def plume_2d(model, time=None, relative_concentration=False, animate=False, **kwargs):
    """Plot contaminant plume as a 2D colormesh, at a specified time.

    Args:
        model : Model object from mibitrans.transport.
        time (float): Point of time for the plot. Will show the closest time step to given value.
            By default, last point in time is plotted.
        relative_concentration (bool, optional) : If set to True, will plot concentrations relative to maximum source
            zone concentrations at t=0. By default, absolute concentrations are shown.
        animate (bool, optional): If True, animation of contaminant plume until given time is shown. Default is
            False.
        **kwargs : Arguments to be passed to plt.pcolormesh().

    Returns a matrix plot of the input plume as object.
    """
    check_model_type(model, allowed_model_types())
    t_pos = check_time_in_domain(model, time)
    if relative_concentration:
        model_concentration = model.relative_cxyt
        z_label = relative_conc_zlabel
    else:
        model_concentration = model.cxyt
        z_label = absolute_conc_zlabel
    # Non animated plot
    if not animate:
        plt.pcolormesh(model.x, model.y, model_concentration[t_pos, :, :], **kwargs)
        plt.xlabel("Distance from source (m)")
        plt.ylabel("Distance from plume center (m)")
        plt.colorbar(label=z_label)
        plot_title = _plot_title_generator("Plume", model, time=model.t[t_pos])
        plt.title(plot_title)

    # Animated plot
    else:
        fig, ax = plt.subplots()
        mesh = ax.pcolormesh(
            model.x, model.y, model_concentration[0, :, :], vmin=0, vmax=np.max(model_concentration), **kwargs
        )
        cbar = fig.colorbar(mesh, ax=ax)
        cbar.set_label(z_label)
        ax.set_xlabel("Distance from source (m)")
        ax.set_ylabel("Distance from plume center (m)")

        def update(frame):
            mesh.set_array(model_concentration[frame, :, :])
            ax.set_title(f"Concentration distribution at t={model.t[frame]} days")
            return mesh

        ani = animation.FuncAnimation(fig=fig, func=update, frames=t_pos + 1)
        return ani


def plume_3d(model, time=None, relative_concentration=False, animate=False, **kwargs):
    """Plot contaminant plume as a 3D surface, at a specified time.

    Args:
        model : Model object from mibitrans.transport.
        time (float): Point of time for the plot. Will show the closest time step to given value.
            By default, last point in time is plotted.
        relative_concentration (bool, optional) : If set to True, will plot concentrations relative to maximum source
            zone concentrations at t=0. By default, absolute concentrations are shown.
        animate (bool, optional): If True, animation of contaminant plume until given time is shown. Default is
            False.
        **kwargs : Arguments to be passed to plt.plot_surface().

    Returns:
        ax (matplotlib.axes._axes.Axes) : Returns matplotlib axes object of plume plot.
    """
    check_model_type(model, allowed_model_types())
    t_pos = check_time_in_domain(model, time)
    if relative_concentration:
        model_concentration = model.relative_cxyt
        z_label = relative_conc_zlabel
    else:
        model_concentration = model.cxyt
        z_label = absolute_conc_zlabel

    # Non animated plot
    xxx = np.tile(model.x, (len(model.t), len(model.y), 1))
    yyy = np.tile(model.y[:, None], (len(model.t), 1, len(model.x)))
    if not animate:
        fig, ax = plt.subplots(subplot_kw={"projection": "3d"})
        ax.plot_surface(xxx[t_pos, :, :], yyy[t_pos, :, :], model_concentration[t_pos, :, :], **kwargs)
        ax.view_init(elev=30, azim=310)
        ax.set_xlabel("Distance from source (m)")
        ax.set_ylabel("Distance from plume center (m)")
        ax.set_zlabel(z_label)
        plot_title = _plot_title_generator("Plume", model, time=model.t[t_pos])
        ax.set_title(plot_title)
        return ax

    # Animated plot
    else:
        if "cmap" not in kwargs and "color" not in kwargs:
            kwargs["color"] = "tab:blue"
        model_max = np.max(model_concentration)
        fig, ax = plt.subplots(subplot_kw={"projection": "3d"})
        surface = ax.plot_surface(
            xxx[0, :, :],
            yyy[0, :, :],
            model_concentration[0, :, :],
            vmin=0,
            vmax=model_max,
            **kwargs,
        )
        ax.set_xlabel("Distance from source (m)")
        ax.set_ylabel("Distance from plume center (m)")
        ax.set_zlabel(z_label)
        ax.set_zlim(0, model_max)

        # plot_surface creates a static surface; need to create new plot every time step
        def update(frame):
            # nonlocal needed in order for the previous plot to be removed before new one is plotted
            nonlocal surface
            surface.remove()
            surface = ax.plot_surface(
                xxx[frame, :, :],
                yyy[frame, :, :],
                model_concentration[frame, :, :],
                vmin=0,
                vmax=model_max,
                **kwargs,
            )
            ax.set_title(f"Concentration distribution at t={model.t[frame]} days")
            return surface

        ani = animation.FuncAnimation(fig=fig, func=update, frames=t_pos + 1)
        return ani



import matplotlib.pyplot as plt
import numpy as np
from matplotlib import animation


def animate_1d(
    x_axis_parameter: np.ndarray, y_axis_parameter, time_parameter, y_names=None, y_colors=None, linestyle=None
):
    """Animate any type and number of input 1D plottables."""
    if not isinstance(y_axis_parameter, list):
        y_axis_parameter = [y_axis_parameter]
    fig, ax = plt.subplots()
    plot_bin = []
    max_conc = 0
    for i, y in enumerate(y_axis_parameter):
        line = ax.plot(x_axis_parameter, y[0, :], color=y_colors[i], label=y_names[i], linestyle=linestyle[i])[0]
        if np.max(y) > max_conc:
            max_conc = np.max(y)
        plot_bin.append(line)
    ax.set_ylim(bottom=0, top=max_conc + max_conc / 10)
    ax.set_xlabel("Distance from source [m]")
    ax.set_ylabel("Concentration [g/m3]")
    ax.legend()
    n_frames = len(time_parameter)

    def update(frame):
        """Update plot with values for the next time step in the animation."""
        for i, y in enumerate(y_axis_parameter):
            plot_bin[i].set_xdata(x_axis_parameter)
            plot_bin[i].set_ydata(y[frame, :])
        ax.set_title(f"Concentration distribution at t={time_parameter[frame]} days")
        return plot_bin

    ani = animation.FuncAnimation(fig=fig, func=update, frames=n_frames)
    return ani



import matplotlib
import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d
import pytest
from mibitrans.data.check_input import DomainValueError
from mibitrans.transport.models import Mibitrans
from mibitrans.visualize.plot_line import breakthrough
from mibitrans.visualize.plot_line import centerline
from mibitrans.visualize.plot_line import transverse
from mibitrans.visualize.plot_surface import plume_2d
from mibitrans.visualize.plot_surface import plume_3d

matplotlib.use("Agg")  # Fixes tkinter.TclError in local tests


@pytest.mark.parametrize(
    "animate, expected",
    [
        (False, matplotlib.axes._axes.Axes),
        (True, matplotlib.animation.FuncAnimation),
    ],
)
# Ignore warning for animation being deleted without being shown, as behaviour is intentional for testing purposes.
@pytest.mark.filterwarnings("ignore:Animation was deleted")
def test_centerline(animate, expected, test_anatrans_model_nodecay, test_anatrans_model_lineardecay):
    """Test if plot object is generated in centerline function."""
    model_nd, results_nd = test_anatrans_model_nodecay
    model_lin, results_lin = test_anatrans_model_lineardecay
    if not animate:
        centerline(results_nd, animate=animate)
        assert isinstance(plt.gca(), expected)
        plt.clf()
        centerline(
            [results_nd, results_lin],
            legend_names=["no decay", "linear decay"],
            animate=animate,
        )
        assert isinstance(plt.gca(), expected)

    else:
        ani = centerline(results_nd, animate=animate)
        assert isinstance(ani, expected)
        anim = centerline(
            [results_nd, results_lin],
            legend_names=["no decay", "linear decay"],
            animate=animate,
        )
        assert isinstance(anim, expected)


@pytest.mark.parametrize(
    "y_pos, time, expected",
    [
        (1, 365, matplotlib.axes._axes.Axes),
        (None, 365, matplotlib.axes._axes.Axes),
        (-1, None, matplotlib.axes._axes.Axes),
        (100000, 365, UserWarning),
        (1, 40000000, UserWarning),
        ("nonsense", 365, TypeError),
        (1, -1, DomainValueError),
    ],
)
def test_parameter_check_centerline(y_pos, time, expected, test_anatrans_model_nodecay):
    """Test if centerline function properly raises warnings and errors for parameters outside domain."""
    model, results = test_anatrans_model_nodecay
    if isinstance(expected, matplotlib.axes._axes.Axes):
        centerline(results, y_pos, time)
        assert isinstance(plt.gca(), expected)
    elif expected is UserWarning:
        with pytest.warns(expected):
            centerline(results, y_pos, time)
            assert isinstance(plt.gca(), matplotlib.axes._axes.Axes)
    elif expected is TypeError or expected is DomainValueError:
        with pytest.raises(expected):
            centerline(results, y_pos, time)


@pytest.mark.parametrize(
    "animate, expected",
    [
        (False, matplotlib.axes._axes.Axes),
        (True, matplotlib.animation.FuncAnimation),
    ],
)
@pytest.mark.filterwarnings("ignore:Animation was deleted")
def test_transverse(animate, expected, test_anatrans_model_nodecay, test_anatrans_model_lineardecay):
    """Test if plot object is generated in transverse function."""
    model_nd, results_nd = test_anatrans_model_nodecay
    model_lin, results_lin = test_anatrans_model_lineardecay
    if not animate:
        transverse(results_nd, x_position=10, animate=animate)
        assert isinstance(plt.gca(), expected)
        plt.clf()
        transverse(
            [results_nd, results_lin],
            x_position=10,
            legend_names=["no decay", "linear decay"],
            animate=animate,
        )
        assert isinstance(plt.gca(), expected)

    else:
        ani = transverse(results_nd, x_position=10, animate=animate)
        assert isinstance(ani, expected)
        anim = transverse(
            [results_nd, results_lin],
            x_position=10,
            legend_names=["no decay", "linear decay"],
            animate=animate,
        )
        assert isinstance(anim, expected)


@pytest.mark.parametrize(
    "x_pos, time, expected",
    [
        (10, 365, matplotlib.axes._axes.Axes),
        (-1, None, matplotlib.axes._axes.Axes),
        (1000000, 365, UserWarning),
        (10, 3000000, UserWarning),
        ("nonsense", 365, TypeError),
        (1, -1, DomainValueError),
    ],
)
def test_parameter_check_transverse(x_pos, time, expected, test_anatrans_model_nodecay):
    """Test if transverse function properly raises warnings and errors for parameters outside domain."""
    model, results = test_anatrans_model_nodecay
    if isinstance(expected, matplotlib.axes._axes.Axes):
        transverse(results, x_pos, time)
        assert isinstance(plt.gca(), expected)
    elif expected is UserWarning:
        with pytest.warns(expected):
            transverse(results, x_pos, time)
            assert isinstance(plt.gca(), matplotlib.axes._axes.Axes)
    elif expected is TypeError or expected is DomainValueError:
        with pytest.raises(expected):
            transverse(results, x_pos, time)


@pytest.mark.parametrize(
    "animate, expected",
    [
        (False, matplotlib.axes._axes.Axes),
        (True, matplotlib.animation.FuncAnimation),
    ],
)
@pytest.mark.filterwarnings("ignore:Animation was deleted")
def test_breakthrough(animate, expected, test_anatrans_model_nodecay, test_anatrans_model_lineardecay):
    """Test if plot object is generated in breakthrough function."""
    model_nd, results_nd = test_anatrans_model_nodecay
    model_lin, results_lin = test_anatrans_model_lineardecay
    if not animate:
        breakthrough(results_nd, x_position=10, animate=animate)
        assert isinstance(plt.gca(), expected)
        plt.clf()
        breakthrough(
            [results_nd, results_lin],
            x_position=10,
            legend_names=["no decay", "linear decay"],
            animate=animate,
        )
        assert isinstance(plt.gca(), expected)
    else:
        ani = breakthrough(results_nd, x_position=10, animate=animate)
        assert isinstance(ani, expected)
        anim = breakthrough(
            [results_nd, results_lin],
            x_position=10,
            legend_names=["no decay", "linear decay"],
            animate=animate,
        )
        assert isinstance(anim, expected)


@pytest.mark.parametrize(
    "x_pos, y_pos, expected",
    [
        (10, 1, matplotlib.axes._axes.Axes),
        (-1, None, matplotlib.axes._axes.Axes),
        (10000000, 365, UserWarning),
        (10, 200000000, UserWarning),
        ("nonsense", 365, TypeError),
        (-10, 1, DomainValueError),
    ],
)
def test_parameter_check_breakthrough(x_pos, y_pos, expected, test_anatrans_model_nodecay):
    """Test if breakthrough function properly raises warnings and errors for parameters outside domain."""
    model, results = test_anatrans_model_nodecay
    if isinstance(expected, matplotlib.axes._axes.Axes):
        breakthrough(results, x_pos, y_pos)
        assert isinstance(plt.gca(), expected)
    elif expected is UserWarning:
        with pytest.warns(expected):
            breakthrough(results, x_pos, y_pos)
            assert isinstance(plt.gca(), matplotlib.axes._axes.Axes)
    elif expected is TypeError or expected is DomainValueError:
        with pytest.raises(expected):
            breakthrough(results, x_pos, y_pos)


@pytest.mark.parametrize(
    "animate, expected",
    [
        (False, matplotlib.axes._axes.Axes),
        (True, matplotlib.animation.FuncAnimation),
    ],
)
@pytest.mark.filterwarnings("ignore:Animation was deleted")
def test_plume_2d(animate, expected, test_anatrans_model_nodecay):
    """Test if plot object is generated in plume 2d function."""
    model, results = test_anatrans_model_nodecay
    if not animate:
        plume_2d(results, animate=animate)
        assert isinstance(plt.gca(), expected)
    else:
        ani = plume_2d(results, animate=animate)
        assert isinstance(ani, expected)


@pytest.mark.parametrize(
    "animate, expected",
    [
        (False, mpl_toolkits.mplot3d.axes3d.Axes3D),
        (True, matplotlib.animation.FuncAnimation),
    ],
)
@pytest.mark.filterwarnings("ignore:Animation was deleted")
def test_plume_3d(animate, expected, test_anatrans_model_nodecay):
    """Test if plot object is generated in plume 3d function."""
    model, results = test_anatrans_model_nodecay
    ax = plume_3d(results, animate=animate)
    assert isinstance(ax, expected)


def test_source_zone(test_source_pars):
    """Test if plot object is generated in source zone function."""
    test_source_pars.visualize()
    assert isinstance(plt.gca(), matplotlib.axes._axes.Axes)


@pytest.mark.parametrize(
    "plottable, expected",
    [
        (-1, TypeError),
        ("test_hydro_pars", TypeError),
    ],
)
def test_input_plotting(plottable, expected, request):
    """Test if input validation plotting function."""
    if isinstance(plottable, str):
        plottable = request.getfixturevalue(plottable)
    with pytest.raises(expected):
        centerline(plottable)
    with pytest.raises(expected):
        transverse(plottable, x_position=1)
    with pytest.raises(expected):
        breakthrough(plottable, x_position=1)
    with pytest.raises(expected):
        plume_2d(plottable)
    with pytest.raises(expected):
        plume_3d(plottable)


@pytest.mark.parametrize(
    "plotter, expected",
    [
        ("centerline", matplotlib.axes._axes.Axes),
        ("transverse", matplotlib.axes._axes.Axes),
        ("breakthrough", matplotlib.axes._axes.Axes),
        ("plume_2d", matplotlib.axes._axes.Axes),
        ("plume_3d", matplotlib.axes._axes.Axes),
        ("plume_3d", mpl_toolkits.mplot3d.axes3d.Axes3D),
    ],
)
def test_innate_plotting_methods(plotter, expected, test_hydro_pars, test_att_pars, test_source_pars, test_model_pars):
    """Test if the plotting class methods of the results object work."""
    model = Mibitrans(test_hydro_pars, test_att_pars, test_source_pars, test_model_pars)
    results = model.run()
    match plotter:
        case "centerline":
            results.centerline()
        case "transverse":
            results.transverse(1)
        case "breakthrough":
            results.breakthrough(1)
        case "plume_2d":
            results.plume_2d()
        case "plume_3d":
            ax = results.plume_3d()
            assert isinstance(ax, expected)
    assert isinstance(plt.gca(), expected)
    plt.clf()



"""Author: Jorrit Bakker.

Module plotting a 3D matrix of contaminant plume concentrations as a line.
"""

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import animation
import mibitrans
from mibitrans.data.check_input import check_model_type
from mibitrans.data.check_input import check_time_in_domain
from mibitrans.data.check_input import check_x_in_domain
from mibitrans.data.check_input import check_y_in_domain

relative_conc_ylabel = r"Relative concentration ($C/C_0$)"
absolute_conc_ylabel = r"Concentration [g/$m^{3}$]"


def allowed_model_types():
    """Return object of parent class that is allowed for input/output."""
    return mibitrans.transport.model_parent.Results


def centerline(
    model, y_position=0, time=None, relative_concentration=False, legend_names=None, animate=False, **kwargs
):
    """Plot center of contaminant plume of one or multiple models as a line, at a specified time and y position.

    Args:
        model : Model object from mibitrans.transport, or list of model objects.
        y_position (float, optional): y-position across the plume (transverse horizontal direction) for the plot.
            By default, the center of the plume at y=0 is plotted.
        time (float, optional): Point of time for the plot. Will show the closest time step to given value.
            By default, last point in time is plotted.
        relative_concentration (bool, optional) : If set to True, will plot concentrations relative to maximum source
            zone concentrations at t=0. By default, absolute concentrations are shown.
        legend_names (str | list, optional): List of legend names as strings, in the same order as given models.
            By default, no legend is shown.
        animate (bool, optional): If True, animation of contaminant plume until given time is shown. If multiple models
            are given as input, dt should be the same for each one to ensure accurate animation. Default is False.
        **kwargs : Arguments to be passed to plt.plot().

    """
    if not isinstance(model, list):
        model = [model]
    if not isinstance(legend_names, list) and legend_names is not None:
        legend_names = [legend_names]

    plot_array_list = []
    # Checks for list model input: dt should be equal, time should be smaller than the smallest end time, y_position
    # should be inside narrowest domain boundaries
    for mod in model:
        check_model_type(mod, allowed_model_types())
        y_pos = check_y_in_domain(mod, y_position)
        t_pos = check_time_in_domain(mod, time)

        if relative_concentration:
            if animate:
                plot_array_list.append(mod.relative_cxyt[:, y_pos, :])
            else:
                plot_array_list.append(mod.relative_cxyt[t_pos, y_pos, :])
            y_label = relative_conc_ylabel
        else:
            if animate:
                plot_array_list.append(mod.cxyt[:, y_pos, :])
            else:
                plot_array_list.append(mod.cxyt[t_pos, y_pos, :])
            y_label = absolute_conc_ylabel

    # Non-animated plot
    if not animate:
        for i, mod in enumerate(model):
            if legend_names is not None:
                plt.plot(mod.x, plot_array_list[i], label=legend_names[i], **kwargs)
            else:
                plt.plot(mod.x, plot_array_list[i], **kwargs)

        plt.ylim(bottom=0)
        plt.xlabel("Distance from source [m]")
        plt.ylabel(y_label)

        plot_title = _plot_title_generator(
            "Centerline", model[0], time=model[0].t[t_pos], y_position=y_position, multiple=len(model) > 1
        )
        plt.title(plot_title)
        if legend_names is not None:
            plt.legend()

    # Animated plot
    else:
        fig, ax = plt.subplots()
        plot_bin = []
        for i, mod in enumerate(model):
            if legend_names is not None:
                line = ax.plot(mod.x, plot_array_list[i][0, :], label=legend_names[i], **kwargs)[0]
            else:
                line = ax.plot(mod.x, plot_array_list[i][0, :], **kwargs)[0]
            plot_bin.append(line)
        ax.set_ylim(bottom=0)
        ax.set_xlabel("Distance from source [m]")
        ax.set_ylabel(y_label)

        if legend_names is not None:
            ax.legend()

        def update(frame):
            for i, mod in enumerate(model):
                plot_bin[i].set_xdata(mod.x)
                plot_bin[i].set_ydata(plot_array_list[i][frame, :])
                ax.set_title(f"Concentration distribution at t={mod.t[frame]} days")
            return plot_bin

        ani = animation.FuncAnimation(fig=fig, func=update, frames=t_pos + 1)
        return ani


def transverse(model, x_position, time=None, relative_concentration=False, legend_names=None, animate=False, **kwargs):
    """Plot concentration distribution as a line horizontal transverse to the plume extent.

    Args:
        model : Model object from mibitrans.transport, or list of model objects.
        x_position : x-position along the plume (longitudinal direction) for the plot.
        time (float): Point of time for the plot. Will show the closest time step to given value.
            By default, last point in time is plotted.
        relative_concentration (bool, optional) : If set to True, will plot concentrations relative to maximum source
            zone concentrations at t=0. By default, absolute concentrations are shown.
        legend_names (str | list, optional): List of legend names as strings, in the same order as given models.
            By default, no legend is shown.
        animate (bool, optional): If True, animation of contaminant plume until given time is shown. If multiple models
            are given as input, dt should be the same for each one to ensure accurate animation. Default is False.
        **kwargs : Arguments to be passed to plt.plot().
    """
    if not isinstance(model, list):
        model = [model]
    if not isinstance(legend_names, list) and legend_names is not None:
        legend_names = [legend_names]

    plot_array_list = []
    # Checks for list model input: dt should be equal, time should be smaller than the smallest end time, y_position
    # should be inside narrowest domain boundaries
    for mod in model:
        check_model_type(mod, allowed_model_types())
        x_pos = check_x_in_domain(mod, x_position)
        t_pos = check_time_in_domain(mod, time)

        if relative_concentration:
            if animate:
                plot_array_list.append(mod.relative_cxyt[:, :, x_pos])
            else:
                plot_array_list.append(mod.relative_cxyt[t_pos, :, x_pos])
            y_label = relative_conc_ylabel
        else:
            if animate:
                plot_array_list.append(mod.cxyt[:, :, x_pos])
            else:
                plot_array_list.append(mod.cxyt[t_pos, :, x_pos])
            y_label = absolute_conc_ylabel

    if not animate:
        for i, mod in enumerate(model):
            if legend_names is not None:
                plt.plot(mod.y, plot_array_list[i], label=legend_names[i], **kwargs)
            else:
                plt.plot(mod.y, plot_array_list[i], **kwargs)

        plt.ylim(bottom=0)
        plt.xlabel("y-position [m]")
        plt.ylabel(y_label)
        plot_title = _plot_title_generator(
            "Transverse", model[0], time=model[0].t[t_pos], x_position=x_position, multiple=len(model) > 1
        )
        plt.title(plot_title)
        if legend_names is not None:
            plt.legend()
    else:
        fig, ax = plt.subplots()
        plot_bin = []
        max_conc = 0
        for i, mod in enumerate(model):
            if legend_names is not None:
                line = ax.plot(mod.y, plot_array_list[i][0, :], label=legend_names[i])[0]
            else:
                line = ax.plot(mod.y, plot_array_list[i][0, :], **kwargs)[0]
            if np.max(plot_array_list[i]) > max_conc:
                max_conc = np.max(plot_array_list[i])
            plot_bin.append(line)
        ax.set_ylim(bottom=0, top=max_conc + max_conc / 10)

        ax.set_xlabel("y-position [m]")
        ax.set_ylabel(y_label)

        if legend_names is not None:
            ax.legend()

        def update(frame):
            for i, mod in enumerate(model):
                plot_bin[i].set_xdata(mod.y)
                plot_bin[i].set_ydata(plot_array_list[i][frame, :])
                ax.set_title(f"Concentration distribution at t={mod.t[frame]} days")
            return plot_bin

        ani = animation.FuncAnimation(fig=fig, func=update, frames=t_pos + 1)
        return ani


def breakthrough(
    model, x_position, y_position=0, relative_concentration=False, legend_names=None, animate=False, **kwargs
):
    """Plot contaminant breakthrough curve at given x and y position in model domain.

    Args:
        model : Model object from mibitrans.transport, or list of model objects.
        x_position : x-position along the plume (longitudinal direction).
        y_position : y-position across the plume (transverse horizontal direction).
            By default, at the center of the plume (at y=0).
        relative_concentration (bool, optional) : If set to True, will plot concentrations relative to maximum source
            zone concentrations at t=0. By default, absolute concentrations are shown.
        legend_names (str | list, optional): List of legend names as strings, in the same order as given models.
            By default, no legend is shown.
        animate (bool, optional): If True, animation of contaminant plume until given time is shown. If multiple models
            are given as input, dt should be the same for each one to ensure accurate animation. Default is False.
        **kwargs : Arguments to be passed to plt.plot().
    """
    if not isinstance(model, list):
        model = [model]
    if not isinstance(legend_names, list) and legend_names is not None:
        legend_names = [legend_names]

    plot_array_list = []
    # Checks for list model input: dt should be equal, time should be smaller than the smallest end time, y_position
    # should be inside narrowest domain boundaries
    for mod in model:
        check_model_type(mod, allowed_model_types())
        x_pos = check_x_in_domain(mod, x_position)
        y_pos = check_y_in_domain(mod, y_position)
        if relative_concentration:
            plot_array_list.append(mod.relative_cxyt[:, y_pos, x_pos])
            y_label = relative_conc_ylabel
        else:
            plot_array_list.append(mod.cxyt[:, y_pos, x_pos])
            y_label = absolute_conc_ylabel

    # Non animated plot
    if not animate:
        for i, mod in enumerate(model):
            if legend_names is not None:
                plt.plot(mod.t, plot_array_list[i], label=legend_names[i], **kwargs)
            else:
                plt.plot(mod.t, plot_array_list[i], **kwargs)

        plt.ylim(bottom=0)
        plt.xlabel("Time [days]")
        plt.ylabel(y_label)
        plot_title = _plot_title_generator(
            "Breakthrough", model[0], x_position=x_position, y_position=y_position, multiple=len(model) > 1
        )
        plt.title(plot_title)
        if legend_names is not None:
            plt.legend()

    # Animated plot
    else:
        fig, ax = plt.subplots()
        plot_bin = []
        max_conc = 0
        max_time = 0
        for i, mod in enumerate(model):
            if legend_names is not None:
                line = ax.plot(mod.t[0], plot_array_list[i][0], label=legend_names[i])[0]
            else:
                line = ax.plot(mod.t[0], plot_array_list[i][0], **kwargs)[0]

            # As plot extent is decided by first initiation of plot, ensure that axis concentration and time limits
            # are corresponding with their maximum values
            if mod.t[-1] > max_time:
                max_time = mod.t[-1]
            if np.max(plot_array_list[i]) > max_conc:
                max_conc = np.max(plot_array_list[i])
            plot_bin.append(line)
        ax.set_xlim(right=max_time)
        ax.set_ylim(bottom=0, top=max_conc + max_conc / 10)
        ax.set_xlabel("Time [days]")
        ax.set_ylabel(y_label)
        if legend_names is not None:
            ax.legend()

        def update(frame):
            for i, mod in enumerate(model):
                plot_bin[i].set_xdata(mod.t[:frame])
                plot_bin[i].set_ydata(plot_array_list[i][:frame])
                ax.set_title(f"Breakthrough curve at t={mod.t[frame]} days")
            return plot_bin

        ani = animation.FuncAnimation(fig=fig, func=update, frames=len(model[0].t))
        return ani


def _plot_title_generator(plot_type, model, time=None, x_position=None, y_position=None, multiple=False):
    if multiple:
        title = f"{plot_type} plot of multiple models, at "
    else:
        title = f"{plot_type} plot of {model.short_description} model, at "

    if time is not None:
        title += f"t = {time} days"
    if x_position is not None:
        if time is not None:
            title += ", "
        title += f"x = {x_position} m"
    if y_position is not None and y_position != 0:
        if time is not None or x_position is not None:
            title += ", "
        title += f"y = {y_position} m"
    title += "."
    return title



import pytest
from mibitrans.data.check_input import DomainValueError
from tests.test_example_data import testingdata_instantreaction_bioscreen
from tests.test_example_data import testingdata_lineardecay_bioscreen
from tests.test_example_data import testingdata_nodecay_bioscreen


@pytest.mark.parametrize(
    "model, expected",
    [
        ("test_bioscreen_model_nodecay", testingdata_nodecay_bioscreen),
        ("test_bioscreen_model_lineardecay", testingdata_lineardecay_bioscreen),
        ("test_bioscreen_model_instantreaction", testingdata_instantreaction_bioscreen),
    ],
)
@pytest.mark.filterwarnings("ignore:Decay rate was set")
def test_transport_equation_numerical_bioscreen(model, expected, request):
    """Test numerical output of transport equation of Anatrans, by comparing to pre-calculated values."""
    model, results = request.getfixturevalue(model)
    assert model.cxyt == pytest.approx(expected)
    assert results.cxyt == pytest.approx(expected)


@pytest.mark.parametrize(
    "x, y, t, expected",
    [
        (9, 0, 629, 1.2260728205395477),
        (15, -7, 256, 0.21033402922523056),
        (-16, 0, 393, DomainValueError),
        ("nonsense", 0, 393, TypeError),
        (16, "nonsense", 393, TypeError),
        (16, 0, -10, DomainValueError),
        (16, 0, "nonsense", TypeError),
    ],
)
def test_bioscreen_sample_linear(x, y, t, expected, test_bioscreen_model_lineardecay):
    """Test if sample method from Bioscreen works correctly, and gives expected output for linear models."""
    model, results = test_bioscreen_model_lineardecay
    if isinstance(expected, float):
        assert model.sample(x, y, t) == pytest.approx(expected)
    elif expected is ValueError or expected is TypeError:
        with pytest.raises(expected):
            model.sample(x, y, t)


@pytest.mark.parametrize(
    "x, y, t, expected",
    [
        (13, 0, 354, 2.721953070355462),
        (11, 3, 752, 5.01432266465888),
        (-16, 0, 393, DomainValueError),
        ("nonsense", 0, 393, TypeError),
        (16, "nonsense", 393, TypeError),
        (16, 0, -10, DomainValueError),
        (16, 0, "nonsense", TypeError),
    ],
)
def test_bioscreen_sample_instantreaction(x, y, t, expected, test_bioscreen_model_instantreaction):
    """Test if sample method from Bioscreen works correctly, and gives expected output for instant reaction models."""
    model, results = test_bioscreen_model_instantreaction
    if isinstance(expected, float):
        assert model.sample(x, y, t) == pytest.approx(expected)
    elif expected is ValueError or expected is TypeError:
        with pytest.raises(expected):
            model.sample(x, y, t)



"""Author: Jorrit Bakker.

File testing functionality of mass_balance module.
"""

import numpy as np
import pytest
from mibitrans.data.parameters import ModelParameters
from mibitrans.data.parameters import SourceParameters
from mibitrans.transport.models import Anatrans
from mibitrans.transport.models import Bioscreen
from mibitrans.transport.models import Mibitrans
from tests.test_example_data import testing_massbalance_instant_ana_inf
from tests.test_example_data import testing_massbalance_instant_mbt
from tests.test_example_data import testing_massbalance_instant_mbt_inf
from tests.test_example_data import testing_massbalance_lineardecay_ana
from tests.test_example_data import testing_massbalance_nodecay_bio


@pytest.fixture(scope="module")
def test_model_pars():
    """ModelParameters fixture with increased spatial resolution, specifically for testing mass balance."""
    return ModelParameters(model_length=50, model_width=30, model_time=3 * 365, dx=1, dy=1, dt=1 * 365)


@pytest.fixture(scope="module")
def test_source_pars_inf():
    """SourceParameters fixture with example data for tests."""
    return SourceParameters(
        source_zone_boundary=np.array([5, 10, 15]),
        source_zone_concentration=np.array([10, 5, 2]),
        depth=10,
        total_mass="inf",
    )


@pytest.fixture(scope="module")
def test_bioscreen_nodecay_model_mb(test_hydro_pars, test_att_pars_nodecay, test_source_pars, test_model_pars):
    """Bioscreen with linear decay fixture mass balance object for testing."""
    obj = Bioscreen(test_hydro_pars, test_att_pars_nodecay, test_source_pars, test_model_pars)
    res = obj.run()
    return res.mass_balance()


@pytest.fixture(scope="module")
def test_anatrans_lineardecay_model_mb(test_hydro_pars, test_att_pars, test_source_pars, test_model_pars):
    """Anatrans with linear decay fixture mass balance object for testing."""
    obj = Anatrans(test_hydro_pars, test_att_pars, test_source_pars, test_model_pars)
    res = obj.run()
    return res.mass_balance()


@pytest.fixture(scope="module")
def test_mibitrans_instantreaction_model_mb(test_hydro_pars, test_att_pars, test_source_pars, test_model_pars):
    """Mibitrans with instant reaction fixture mass balance object for testing."""
    obj = Mibitrans(test_hydro_pars, test_att_pars, test_source_pars, test_model_pars)
    obj.instant_reaction(dict(delta_oxygen=0.5, delta_nitrate=0.5, ferrous_iron=0.5, delta_sulfate=0.5, methane=0.5))
    res = obj.run()
    return res.mass_balance()


@pytest.fixture(scope="module")
def test_mibitrans_instantreaction_model_mb_inf(test_hydro_pars, test_att_pars, test_source_pars_inf, test_model_pars):
    """Mibitrans with instant reaction and infinite source mass fixture mass balance  object for testing."""
    obj = Mibitrans(test_hydro_pars, test_att_pars, test_source_pars_inf, test_model_pars)
    obj.instant_reaction(dict(delta_oxygen=0.5, delta_nitrate=0.5, ferrous_iron=0.5, delta_sulfate=0.5, methane=0.5))
    res = obj.run()
    return res.mass_balance()


@pytest.fixture(scope="module")
def test_anatrans_instantreaction_model_mb_inf(test_hydro_pars, test_att_pars, test_source_pars_inf, test_model_pars):
    """Anatrans with instant reaction and infinite source mass fixture mass balance  object for testing."""
    obj = Anatrans(test_hydro_pars, test_att_pars, test_source_pars_inf, test_model_pars)
    obj.instant_reaction(dict(delta_oxygen=0.5, delta_nitrate=0.5, ferrous_iron=0.5, delta_sulfate=0.5, methane=0.5))
    res = obj.run()
    return res.mass_balance()


@pytest.mark.parametrize(
    "model, expected",
    [
        ("test_bioscreen_nodecay_model_mb", testing_massbalance_nodecay_bio),
        ("test_anatrans_lineardecay_model_mb", testing_massbalance_lineardecay_ana),
        ("test_mibitrans_instantreaction_model_mb", testing_massbalance_instant_mbt),
        ("test_mibitrans_instantreaction_model_mb_inf", testing_massbalance_instant_mbt_inf),
        ("test_anatrans_instantreaction_model_mb_inf", testing_massbalance_instant_ana_inf),
    ],
)
@pytest.mark.filterwarnings("ignore:Decay rate was set")
@pytest.mark.filterwarnings("ignore:Contaminant plume extents")
def test_balance_numerical_mibitrans(model, expected, request) -> None:
    """Test if mass balance is correctly calculated by comparing to precomputed results for Mibitrans model."""
    mb_object = request.getfixturevalue(model)
    dictionary = mb_object.__dict__
    for key, output_item in expected.items():
        if isinstance(output_item, dict):
            for key_2, value_2 in output_item.items():
                assert value_2 == pytest.approx(dictionary[key][key_2])
        else:
            assert output_item == pytest.approx(dictionary[key]), f"mb_item {key}"

    mb_object(2 * 365)
    dictionary_single = mb_object.__dict__
    for key, output_item in expected.items():
        if isinstance(output_item, np.ndarray):
            assert output_item[1] == pytest.approx(dictionary_single[key]), f"mb_item {key}"
        elif isinstance(output_item, dict):
            for key_2, value_2 in output_item.items():
                assert value_2[1] == pytest.approx(dictionary_single[key][key_2]), f"mb_item {key}"
        else:
            assert output_item == pytest.approx(dictionary_single[key]), f"mb_item {key}"



"""Author: Jorrit Bakker.

Module including various methods to visualize (input) parameter conditions, intended to only be called internally.
"""

import matplotlib
import matplotlib.pyplot as plt
import numpy as np


def source_zone(source_parameters):
    """Visualize source zone conditions."""
    source_y = source_parameters.source_zone_boundary
    source_c = source_parameters.source_zone_concentration

    y_discretization = np.linspace(-source_y[-1] - source_y[-1] / 10, source_y[-1] + source_y[-1] / 10, 10000)
    c_values = np.zeros(len(y_discretization))
    for i, y in enumerate(source_y[::-1]):
        c_values = np.where((y_discretization <= y) & (y_discretization >= -y), source_c[-(i + 1)], c_values)

    indexer = np.linspace(1, 0.3, len(source_y))
    colormap = matplotlib.colormaps["YlGnBu"]

    plt.figure(dpi=300)

    for i, y in enumerate(source_y):
        if i == 0:
            plt.fill_betweenx(
                y=y_discretization,
                x1=c_values,
                where=(y_discretization <= y) & (y_discretization >= -y),
                color=colormap(indexer[i]),
                zorder=len(source_y) + 2,
            )
        else:
            plt.fill_betweenx(
                y=y_discretization,
                x1=c_values,
                # Boolean array for domain of source zone i
                where=((y_discretization <= y) & (y_discretization > source_y[i - 1]))
                | ((y_discretization >= -y) & (y_discretization < -source_y[i - 1])),
                color=colormap(indexer[i]),
                zorder=len(source_y) + 2 - i,
            )

    plt.xlabel(r"Source zone concentration $g/m^3$")
    plt.ylabel("Source zone y-coordinate")
    plt.title("Concentration distribution in the source zone")


def model_grid(model_parameters):
    """Visualize the model grid."""
    return None



"""Author: Jorrit Bakker.

Module handling testing of data input functionality
"""

import matplotlib.pyplot as plt
import numpy as np
import pytest
from mibitrans.data.check_input import DomainValueError
from mibitrans.data.check_input import MissingValueError
from mibitrans.data.parameter_information import UtilizationFactor
from mibitrans.data.parameters import AttenuationParameters
from mibitrans.data.parameters import HydrologicalParameters
from mibitrans.data.parameters import ModelParameters
from mibitrans.data.parameters import SourceParameters


# Test HydrologicalParameters
@pytest.mark.parametrize(
    "parameters, error",
    [
        (dict(velocity=1, porosity=0.2, alpha_x=1, alpha_y=1), None),
        (dict(h_gradient=1, h_conductivity=1, porosity=0.2, alpha_x=1, alpha_y=1), None),
        (dict(), MissingValueError),
        (dict(porosity=0.2, alpha_x=1, alpha_y=1), MissingValueError),
        (dict(velocity=1, alpha_x=1, alpha_y=1), MissingValueError),
        (dict(velocity=1, porosity=0.2, alpha_y=1), MissingValueError),
        (dict(velocity=1, porosity=0.2, alpha_x=1), MissingValueError),
        (dict(h_gradient=1, porosity=0.2, alpha_x=1, alpha_y=1), MissingValueError),
        (dict(h_conductivity=1, porosity=0.2, alpha_x=1, alpha_y=1), MissingValueError),
        (dict(velocity="1", porosity=0.2, alpha_x=1, alpha_y=1), TypeError),
        (dict(velocity=1, porosity="2", alpha_x=1, alpha_y=1), TypeError),
        (dict(velocity=-1, porosity=0.2, alpha_x=1, alpha_y=1), DomainValueError),
        (dict(velocity=1, porosity=2, alpha_x=1, alpha_y=1), DomainValueError),
        (dict(velocity=1, porosity=0.2, alpha_x=1, alpha_y=1, alpha_z=-1), DomainValueError),
    ],
)
def test_hyrologicalparameters_validation(parameters, error) -> None:
    """Test validation check of HydrologicalParameters dataclass."""
    if error is None:
        HydrologicalParameters(**parameters)
    else:
        with pytest.raises(error):
            HydrologicalParameters(**parameters)


@pytest.mark.parametrize(
    "parameters, value, error",
    [
        (dict(velocity=1, porosity=0.2, alpha_x=1, alpha_y=1), 2, None),
        (dict(velocity=1, porosity=0.2, alpha_x=1, alpha_y=1), "no", TypeError),
        (dict(velocity=1, porosity=0.2, alpha_x=1, alpha_y=1), -1, DomainValueError),
    ],
)
def test_hydrologicalparameters_setattribute(parameters, value, error) -> None:
    """Test validation of parameters after initialization."""
    hydro = HydrologicalParameters(**parameters)
    if error is None:
        hydro.alpha_x = value
    else:
        with pytest.raises(error):
            hydro.alpha_x = value


@pytest.mark.parametrize(
    "test, param, expected",
    [
        (dict(velocity=1, porosity=0.5, alpha_x=2, alpha_y=3), "velocity", 1),
        (dict(velocity=1, porosity=0.5, h_conductivity=2, h_gradient=2, alpha_x=2, alpha_y=3), "velocity", 8),
    ],
)
def test_hyrologicalparameters_output(test, param, expected) -> None:
    """Test output of HydrologicalParameters dataclass."""
    if "velocity" in test.keys() and "h_gradient" in test.keys():
        with pytest.warns(UserWarning):
            hydro = HydrologicalParameters(**test)
    else:
        hydro = HydrologicalParameters(**test)
    assert getattr(hydro, param) == expected


# Test AttenuationParameters
@pytest.mark.parametrize(
    "parameters, error",
    [
        (dict(decay_rate=0.2), None),
        (dict(half_life=2), None),
        (dict(retardation=1), None),
        (dict(diffusion=0.00004), None),
        (dict(decay_rate=0.2, retardation=2, diffusion=0.00002), None),
        (dict(bulk_density=1, partition_coefficient=1, fraction_organic_carbon=1), None),
        (dict(bulk_density=1, partition_coefficient=1, fraction_organic_carbon=1), None),
        (dict(), None),
        (dict(decay_rate=1, half_life=1), UserWarning),
        (dict(half_life="one"), TypeError),
        (dict(half_life=-1), DomainValueError),
        (dict(retardation="one"), TypeError),
        (dict(retardation=0.1), DomainValueError),
        (dict(retardation=1, fraction_organic_carbon="no"), TypeError),
        (dict(retardation=1, fraction_organic_carbon=2), DomainValueError),
    ],
)
def test_attenuationparameters_validation(parameters, error) -> None:
    """Test validation check of AdsorptionParameters dataclass."""
    if error is None:
        AttenuationParameters(**parameters)
    elif error is UserWarning:
        with pytest.warns(error):
            AttenuationParameters(**parameters)
    else:
        with pytest.raises(error):
            AttenuationParameters(**parameters)


@pytest.mark.parametrize(
    "test, param, expected",
    [
        (dict(retardation=1), "retardation", 1),
        (dict(decay_rate=2), "decay_rate", 2),
        (dict(decay_rate=2), "half_life", np.log(2) / 2),
        (dict(half_life=2), "half_life", 2),
        (dict(half_life=2), "decay_rate", np.log(2) / 2),
        (dict(decay_rate=2, half_life=2), "decay_rate", 2),
        (dict(decay_rate=2, half_life=2), "half_life", np.log(2) / 2),
    ],
)
@pytest.mark.filterwarnings("ignore:Both contaminant decay rate")
def test_attenuationparameters_output(test, param, expected) -> None:
    """Test output of AttenuationParameters dataclass."""
    att = AttenuationParameters(**test)
    assert getattr(att, param) == expected


@pytest.mark.parametrize(
    "test, parameter, value, error",
    [
        (dict(retardation=1), "retardation", 1, None),
        (
            dict(bulk_density=1, partition_coefficient=1, fraction_organic_carbon=1),
            "fraction_organic_carbon",
            {},
            TypeError,
        ),
        (
            dict(bulk_density=1, partition_coefficient=1, fraction_organic_carbon=1),
            "fraction_organic_carbon",
            2,
            DomainValueError,
        ),
        (dict(half_life=3), "half_life", 0, None),
    ],
)
def test_attenuationparameters_setattribute(test, value, parameter, error) -> None:
    """Test validation of parameters after initialization."""
    att = AttenuationParameters(**test)
    if error is None:
        setattr(att, parameter, value)
        assert getattr(att, parameter) == value
    else:
        with pytest.raises(error):
            setattr(att, parameter, value)


@pytest.mark.parametrize(
    "test, expected",
    [
        (dict(util_oxygen=1, util_nitrate=2, util_ferrous_iron=3, util_sulfate=4, util_methane=5), None),
        (dict(util_oxygen="1", util_nitrate=2, util_ferrous_iron=3, util_sulfate=4, util_methane=5), TypeError),
        (dict(util_oxygen=-1, util_nitrate=2, util_ferrous_iron=3, util_sulfate=4, util_methane=5), DomainValueError),
    ],
)
def test_attenuationparameters_utilization(test, expected, test_att_pars) -> None:
    """Test set_utilization_factor method of AttenuationParameters dataclass."""
    if expected is None:
        UtilizationFactor(**test)
    else:
        with pytest.raises(expected):
            UtilizationFactor(**test)


# Test SourceParameters
@pytest.mark.parametrize(
    "parameters, error",
    [
        (dict(source_zone_boundary=[1, 2], source_zone_concentration=[3, 2], depth=5), None),
        (dict(source_zone_boundary=[1, 2], source_zone_concentration=[3, 2], depth=5, total_mass=2), None),
        (dict(source_zone_boundary=[1, 2], source_zone_concentration=[3, 2], depth=5, total_mass="inf"), None),
        (dict(source_zone_boundary=[1, 2], source_zone_concentration=[3, 2], depth=5, total_mass="infint"), None),
        (dict(source_zone_boundary=[1, 2], source_zone_concentration=[3, 2], depth=5, total_mass=np.inf), None),
        (dict(source_zone_boundary=1, source_zone_concentration=[3], depth=5, total_mass=2), None),
        (
            dict(source_zone_boundary=np.array([1, 2, 3]), source_zone_concentration=[3, 2, 1], depth=5, total_mass=2),
            None,
        ),
        (dict(), MissingValueError),
        (dict(source_zone_boundary=(1, 2), source_zone_concentration=[3, 2], depth=5, total_mass=2), TypeError),
        (dict(source_zone_boundary=["one", 2], source_zone_concentration=[3, 2], depth=5, total_mass=2), TypeError),
        (
            dict(source_zone_boundary=[1, 2], source_zone_concentration=np.array([-3, 2]), depth=5, total_mass=2),
            DomainValueError,
        ),
        (
            dict(source_zone_boundary=np.array(([1, 2, 3], [4, 5, 6])), source_zone_concentration=[3, 2], depth=5),
            ValueError,
        ),
        (dict(source_zone_boundary=[1, 2], source_zone_concentration=[2, 3], depth=5, total_mass=2), ValueError),
        (dict(source_zone_boundary=[-1, 2], source_zone_concentration=[3, 2], depth=5, total_mass=2), DomainValueError),
        (dict(source_zone_boundary=-1, source_zone_concentration=3), DomainValueError),
        (dict(source_zone_boundary=[1, 2, 3], source_zone_concentration=[3, 2], depth=5, total_mass=2), ValueError),
        (dict(source_zone_boundary=[1, 2], source_zone_concentration=[3, 2], depth="five", total_mass=2), TypeError),
        (dict(source_zone_boundary=[1, 2], source_zone_concentration=[3, 2], depth=-5, total_mass=2), DomainValueError),
        (dict(source_zone_boundary=[1, 2], source_zone_concentration=[3, 2], depth=5, total_mass=[2, 3]), TypeError),
        (dict(source_zone_boundary=[1, 2], source_zone_concentration=[3, 2], depth=5, total_mass=-2), DomainValueError),
        (dict(source_zone_boundary=[1, 2], source_zone_concentration=[3, 2], depth=5, total_mass="nons"), ValueError),
    ],
)
def test_sourceparameters_validation(parameters, error) -> None:
    """Test validation check of SourceParameters dataclass."""
    if error is None:
        SourceParameters(**parameters)
    else:
        with pytest.raises(error):
            SourceParameters(**parameters)


# Test SourceParameters
@pytest.mark.parametrize(
    "parameter, value, error",
    [
        ("source_zone_boundary", [1, 2, 3], None),
        ("source_zone_concentration", [3, 2, 1], None),
        ("total_mass", 1000, None),
        ("total_mass", "infini", None),
        ("total_mass", np.inf, None),
        ("source_zone_concentration", [1, 2, 3], ValueError),
        ("source_zone_concentration", 1, ValueError),
        ("source_zone_concentration", "No", TypeError),
    ],
)
def test_sourceparameters_validation_setattr(parameter, value, error) -> None:
    """Test validation check of SourceParameters dataclass. when setting attributes."""
    src = SourceParameters(
        source_zone_boundary=np.array([1, 2, 3]), source_zone_concentration=np.array([3, 2, 1]), depth=5
    )
    if error is None:
        setattr(src, parameter, value)
    else:
        with pytest.raises(error):
            setattr(src, parameter, value)


@pytest.mark.parametrize(
    "test, param, expected",
    [
        (
            dict(source_zone_boundary=[1, 2, 3], source_zone_concentration=[6, 4, 2], depth=5, total_mass=2),
            "source_zone_boundary",
            np.array([1, 2, 3]),
        ),
        (
            dict(source_zone_boundary=[2, 3, 1], source_zone_concentration=[4, 2, 6], depth=5, total_mass=2),
            "source_zone_boundary",
            np.array([1, 2, 3]),
        ),
        (
            dict(source_zone_boundary=[2, 3, 1], source_zone_concentration=[4, 2, 6], depth=5, total_mass=2),
            "source_zone_concentration",
            np.array([6, 4, 2]),
        ),
        (
            dict(source_zone_boundary=[1, 2, 3], source_zone_concentration=[6, 4, 2], depth=5, total_mass="inf"),
            "total_mass",
            np.inf,
        ),
    ],
)
def test_sourceparameters_output(test, param, expected) -> None:
    """Test output of SourceParameters dataclass."""
    unordered = np.array(test["source_zone_boundary"]) < test["source_zone_boundary"][0]
    if True in unordered:
        with pytest.warns(UserWarning):
            source = SourceParameters(**test)
    else:
        source = SourceParameters(**test)
    assert source.__dict__[param] == pytest.approx(expected)


def test_sourceparameters_visualize():
    """Test if source zone visualization creates a plot."""
    source = SourceParameters(np.array([1, 2, 3]), np.array([3, 2, 1]), 10, 1000)
    source.visualize()
    assert isinstance(plt.gca(), plt.Axes)


# Test ModelParameters
@pytest.mark.parametrize(
    "parameters, error",
    [
        (dict(model_length=1, model_width=1, model_time=1), None),
        (dict(model_length=1, model_width=1, model_time=1, dx=1, dy=1, dt=1), None),
        (dict(dx=1, dy=1, dt=1), TypeError),
        (dict(model_length="one", model_width=1, model_time=1), TypeError),
        (dict(model_length=-2, model_width=1, model_time=1), DomainValueError),
        (dict(model_length=1, model_width=1, model_time=1, dx=2), ValueError),
        (dict(model_length=1, model_width=1, model_time=1, dy=2), ValueError),
        (dict(model_length=1, model_width=1, model_time=1, dt=2), ValueError),
    ],
)
def test_modelparameters_validation(parameters, error) -> None:
    """Test validation check of ModelParameters dataclass."""
    if error is None:
        ModelParameters(**parameters)
    else:
        with pytest.raises(error):
            ModelParameters(**parameters)


@pytest.mark.parametrize(
    "parameter, value, error",
    [
        ("model_length", 3, None),
        ("dy", 0.2, None),
        ("model_time", 3, None),
        ("model_length", "nonsense", TypeError),
        ("model_length", 0.1, ValueError),
        ("dx", 2, ValueError),
        ("model_width", "nonsense", TypeError),
        ("model_width", 0.1, ValueError),
        ("dy", 2, ValueError),
        ("model_time", "nonsense", TypeError),
        ("model_time", 0.1, ValueError),
        ("dt", 2, ValueError),
    ],
)
def test_modelparameters_validation_setattr(parameter, value, error) -> None:
    """Test validation check of ModelParameters dataclass when setting attributes."""
    mod = ModelParameters(model_length=1, model_width=1, model_time=1, dx=0.5, dy=0.5, dt=0.5)
    if error is None:
        setattr(mod, parameter, value)
    else:
        with pytest.raises(error):
            setattr(mod, parameter, value)


@pytest.mark.parametrize(
    "test, param, expected",
    [
        (dict(model_length=1, model_width=1, model_time=1, dx=0.5), "model_length", 1),
        (dict(model_length=1, model_width=1, model_time=1, dx=0.5), "dx", 0.5),
    ],
)
def test_modelparameters_output(test, param, expected) -> None:
    """Test output of ModelParameters dataclass."""
    model = ModelParameters(**test)
    assert model.__dict__[param] == expected


def test_calculation_optional_discretization():
    """Test if model discretization is calculated if not given."""
    model = ModelParameters(model_length=100, model_width=50, model_time=10)
    assert model.dx, "Model dx should have been calculated, but was not."
    assert model.dy, "Model dy should have been calculated, but was not."
    assert model.dt, "Model dt should have been calculated, but was not."



"""Author: Jorrit Bakker.

Module calculating the mass balance based on base parameters.
"""

import warnings
import numpy as np
import mibitrans
from mibitrans.analysis.parameter_calculations import calculate_discharge_and_average_source_zone_concentration
from mibitrans.data.check_input import check_model_type
from mibitrans.data.check_input import check_time_in_domain
from mibitrans.data.check_input import validate_input_values


class MassBalance:
    """Calculate mass balance characteristics of input model."""

    def __init__(self, results, time, verbose=False):
        """Mass balance object with source and plume characteristics at given time(s), of input model.

        Args:
            results: Input model for which mass balance is calculated, should be a child class of Transport3D.
            time (float | str): Time at which to initially calculate the mass balance. Either as a value between 0 and
                model end time. Or as 'all', which will calculate mass balance attributes for each time step as arrays.
            verbose (bool, optional): Verbose mode. Defaults to False.

        Call:
            Calling the MassBalance object will recalculate the mass balance characteristics of input model for given
                input time.

        Properties:
            plume_mass: Mass of the contaminant plume inside the model extent, at the given time(s), in [g].
            source_mass: Mass of the contaminant source at the given time(s), in [g]. No values are given for models
                with infinite source mass.
            delta_source: Difference in mass between contaminant source at given time and source at t = 0, in [g].
            degraded_mass: Mass of plume contaminant degradation at the given time(s), compared to a model without
                degradation, in [g]. Has no value if model does not consider degradation.
            model_without_degradation: Object of model without degradation. Has no value if model does not consider
                degradation.
            instant_reaction_degraded_mass(self): Difference in plume mass instant reaction with and without
                biodegradation capacity subtracted, in [g].
            electron_acceptor_change(self): Change in electron acceptor/byproduct masses at the given time(s), in [g].
                Only for instant reaction.
        """
        check_model_type(results, mibitrans.transport.model_parent.Results)
        self.results = results
        self.verbose = verbose
        self.t = self._time_check(time)

        self._plume_mass_t = None
        self._source_mass_t = None
        self._delta_source_t = None
        self._degraded_mass_t = None
        self._electron_acceptor_change_t = None
        self._instant_reaction_degraded_mass_t = None
        self._model_without_degradation = None

        # Volume of single cell, as dx * dy * source thickness
        self.cellsize = (
            abs(results.x[0] - results.x[1]) * abs(results.y[0] - results.y[1]) * results.source_parameters.depth
        )

        # Mass balance output differs if the source is represented as an infinite mass.
        if self.results.source_parameters.total_mass == np.inf:
            self.source_mass_finite = False
        else:
            self.source_mass_finite = True

        match self.results.mode:
            # Instant reaction model
            case "instant_reaction":
                self.model_instant_reaction = True
                self.model_degradation = True
            # Linear decay model
            case "linear" if self.results.attenuation_parameters.decay_rate > 0:
                self.model_degradation = True
                self.model_instant_reaction = False
            # No decay model
            case _:
                self.model_degradation = False
                self.model_instant_reaction = False

        if self.verbose:
            print("Calculating mass balance...")

        self._calculation_routine()

    def __call__(self, time=None, method=None):
        """Recalculate the mass balance characteristics of input model for given time and method."""
        if time:
            self.t = self._time_check(time)

        if self.verbose:
            print("Recalculating mass balance...")

        self._calculation_routine()

    @property
    def plume_mass(self):
        """Mass of the contaminant plume in the model extent, at the given time(s), in [g]."""
        return self._plume_mass_t

    @property
    def source_mass(self):
        """Mass of the contaminant source at the given time(s), in [g]. No values are given for infinite source mass."""
        return self._source_mass_t

    @property
    def delta_source(self):
        """Difference in mass between contaminant source at given time and source at t = 0, in [g]."""
        return self._delta_source_t

    @property
    def degraded_mass(self):
        """Mass of plume contaminant degradation at the given time(s), compared to a no degradation model, in [g]."""
        return self._degraded_mass_t

    @property
    def model_without_degradation(self):
        """Model with no degradation used to compare with given model."""
        return self._model_without_degradation

    @property
    def instant_reaction_degraded_mass(self):
        """Difference in plume mass instant reaction with and without biodegradation capacity subtracted, in [g].

        For the instant reaction model, the underlying assumption reads that observed concentrations in the source zone
        are post-degradation. Therefore, the source concentrations without any biodegradation would be higher, the
        amount which is determined by the biodegradation capacity. Then, according to this method, the degraded mass
        is the difference between plume mass before and after subtracting the biodegradation capacity.
        """
        return self._instant_reaction_degraded_mass_t

    @property
    def electron_acceptor_change(self):
        """Change in electron acceptor/byproduct masses at the given time(s), in [g]. Only for instant reaction.

        Electron acceptor/byproduct consumption or generation is based on the degraded plume mass (specifically
        'instant_reaction_degraded_mass'), the utilization factor and relative abundance of the acceptors/byproducts.
        Under the governing assumptions of the instant reaction model, a crude estimate of the total consumption of
        electron acceptors and the generation of byproduct is calculated.
        """
        return self._electron_acceptor_change_t

    def source_threshold(self, threshold):
        """Calculate when source mass is below given threshold. No values are given for infinite source mass."""
        validate_input_values("threshold", threshold)
        if not self.source_mass_finite:
            raise ValueError("Source mass is infinite and therefore cannot go below given threshold.")
        else:
            time_to_threshold = (
                -1 / self.results.k_source * np.log(threshold / self.results.source_parameters.total_mass)
            )
        return time_to_threshold

    def _calculation_routine(self):
        """Perform mass_balance calculations."""
        self._check_model_extent()
        self._plume_mass_t = self._calculate_plume_mass(self.results)
        self._source_mass_t = self._calculate_source_mass()
        self._delta_source_t = self._calculate_delta_source()
        if self.model_degradation:
            self._model_without_degradation = self._calculate_model_without_degradation()
            self._degraded_mass_t = self._calculate_degraded_mass()
        if self.model_instant_reaction:
            self._instant_reaction_degraded_mass_t = self._calculate_instant_reaction_degraded_mass()
            self._electron_acceptor_change_t = self._calculate_electron_acceptor_change()

    def _check_model_extent(self):
        """Check if contaminant plume at given time is reasonably situated within the model extent."""
        # Relative concentration considered to be boundary of the plume extent.
        extent_threshold_value = 0.01
        if isinstance(self.t, np.ndarray):
            cxyt_y_boundary = self.results.relative_cxyt[:, [0, -1], :]
            cxyt_x_boundary = self.results.relative_cxyt[:, :, -1]
        else:
            cxyt_y_boundary = self.results.relative_cxyt[self._t_index, [0, -1], :]
            cxyt_x_boundary = self.results.relative_cxyt[self._t_index, :, -1]

        y_boundary_above_threshold = np.where(cxyt_y_boundary > extent_threshold_value, cxyt_y_boundary, 0.0)
        x_boundary_above_threshold = np.where(cxyt_x_boundary > extent_threshold_value, cxyt_x_boundary, 0.0)
        if np.sum(y_boundary_above_threshold) > 0:
            y_max = np.round(
                np.max(y_boundary_above_threshold) * np.max(self.results.source_parameters.source_zone_concentration), 2
            )
            warnings.warn(
                "Contaminant plume extents beyond the model width, with a maximum concentration at the "
                f"boundary of {y_max}g/m3. To ensure reliable mass balance, re-run the model with increased dimensions "
                "to include the entire plume width in the model extent."
            )
        if np.sum(x_boundary_above_threshold) > 0:
            x_max = np.round(
                np.max(x_boundary_above_threshold) * np.max(self.results.source_parameters.source_zone_concentration), 2
            )
            warnings.warn(
                "Contaminant plume extents beyond the model length, with a maximum concentration at the "
                f"boundary of {x_max}g/m3. To ensure reliable mass balance, re-run the model with increased dimensions "
                "to include the entire plume length in the model extent."
            )

    def _calculate_plume_mass(self, model):
        """Calculate plume mass of input model, for the given time(s)."""
        # Plume mass of model; concentration is converted to mass by multiplying by cellsize and pore space.
        if isinstance(self.t, np.ndarray):
            plume_mass_t = np.sum(
                model.cxyt[:, :, 1:] * self.cellsize * self.results.hydrological_parameters.porosity, axis=(1, 2)
            )
        else:
            plume_mass_t = np.sum(
                model.cxyt[self._t_index, :, 1:] * self.cellsize * self.results.hydrological_parameters.porosity
            )

        return plume_mass_t

    def _calculate_source_mass(self):
        """Calculate source mass of input model, for the given time(s)."""
        if self.source_mass_finite:
            source_mass_t = self.results.source_parameters.total_mass * np.exp(-self.results.k_source * self.t)
        else:
            source_mass_t = np.inf

        return source_mass_t

    def _calculate_delta_source(self):
        """Calculate difference in source mass between t=0 and given time(s)."""
        if self.source_mass_finite:
            delta_source_t = self.results.source_parameters.total_mass - self._source_mass_t
        else:
            Q, c0_avg = calculate_discharge_and_average_source_zone_concentration(self.results)
            delta_source_t = Q * c0_avg * self.t
        return delta_source_t

    def _calculate_model_without_degradation(self):
        """Make a no degradation model for comparison, pass the input parameters to a new class instance as kwargs."""
        model_without_degradation = self.results.model_type(**self.results.input_parameters)
        model_without_degradation.attenuation_parameters.decay_rate = 0
        model_without_degradation.run()
        return model_without_degradation

    def _calculate_degraded_mass(self):
        """Calculate difference between input model plume mass and no degradation model, for a given time(s)."""
        no_degradation_plume_mass = self._calculate_plume_mass(self._model_without_degradation)
        degraded_mass_t = no_degradation_plume_mass - self._plume_mass_t
        return degraded_mass_t

    def _calculate_instant_reaction_degraded_mass(self):
        """Calculate difference between input model plume mass and no degradation model for instant reaction model."""
        if isinstance(self.t, np.ndarray):
            plume_mass_t_noBC = np.sum(
                self.results.cxyt_noBC[:, :, 1:] * self.cellsize * self.results.hydrological_parameters.porosity,
                axis=(1, 2),
            )
        else:
            plume_mass_t_noBC = np.sum(
                self.results.cxyt_noBC[self._t_index, :, 1:]
                * self.cellsize
                * self.results.hydrological_parameters.porosity
            )

        degraded_mass_instant_t = plume_mass_t_noBC - self._plume_mass_t
        return degraded_mass_instant_t

    def _calculate_electron_acceptor_change(self):
        """Calculate the change in electron acceptor mass for given time(s) for instant reaction model."""
        mass_fraction_degraded_acceptor = self.results.electron_acceptors.array / self.results.biodegradation_capacity
        electron_acceptor_change = {}
        electron_acceptors = ["oxygen", "nitrate", "ferrous_iron", "sulfate", "methane"]
        for i, ea in enumerate(electron_acceptors):
            electron_acceptor_change[ea] = self._instant_reaction_degraded_mass_t * mass_fraction_degraded_acceptor[i]

        return electron_acceptor_change

    def _time_check(self, time):
        """Check if time input is valid."""
        if time is None or time == "all":
            t = self.results.t
        elif isinstance(time, str):
            warnings.warn("String not recognized, defaulting to 'all', for all time points.")
            t = self.results.t
        else:
            self._t_index = check_time_in_domain(self.results, time)
            t = float(self.results.t[self._t_index])
        return t



import pytest
from mibitrans.data.check_input import DomainValueError
from tests.test_example_data import testingdata_instantreaction_mibitrans
from tests.test_example_data import testingdata_lineardecay_mibitrans
from tests.test_example_data import testingdata_nodecay_mibitrans


@pytest.mark.parametrize(
    "model, expected",
    [
        ("test_mibitrans_model_nodecay", testingdata_nodecay_mibitrans),
        ("test_mibitrans_model_lineardecay", testingdata_lineardecay_mibitrans),
        ("test_mibitrans_model_instantreaction", testingdata_instantreaction_mibitrans),
    ],
)
@pytest.mark.filterwarnings("ignore:Decay rate was set")
def test_transport_equation_numerical_mibitrans(model, expected, request):
    """Test numerical output of transport equation of Mibitrans, by comparing to pre-calculated values."""
    model, results = request.getfixturevalue(model)
    assert model.cxyt == pytest.approx(expected)
    assert results.cxyt == pytest.approx(expected)


@pytest.mark.parametrize(
    "x, y, t, expected",
    [
        (16, 0, 393, 0.2980583920684923),
        (24, -5, 283, 0.03725197246248769),
        (-16, 0, 393, DomainValueError),
        ("nonsense", 0, 393, TypeError),
        (16, "nonsense", 393, TypeError),
        (16, 0, -10, DomainValueError),
        (16, 0, "nonsense", TypeError),
    ],
)
def test_mibitrans_linear_sample(x, y, t, expected, test_mibitrans_model_lineardecay):
    """Test if sample method from Mibitrans works correctly, and gives expected output for linear models."""
    model, results = test_mibitrans_model_lineardecay
    if isinstance(expected, float):
        assert model.sample(x, y, t) == pytest.approx(expected)
    elif expected is ValueError or expected is TypeError:
        with pytest.raises(expected):
            model.sample(x, y, t)


@pytest.mark.parametrize(
    "x, y, t, expected",
    [
        (20, 0, 476, 3.8101869779573443),
        (11, 7, 193, 2.0276832492832924),
        (-16, 0, 393, DomainValueError),
        ("nonsense", 0, 393, TypeError),
        (16, "nonsense", 393, TypeError),
        (16, 0, -10, DomainValueError),
        (16, 0, "nonsense", TypeError),
    ],
)
def test_mibitrans_instant_sample(x, y, t, expected, test_mibitrans_model_instantreaction):
    """Test if sample method from Mibitrans works correctly, and gives expected output for instant reaction models."""
    model, results = test_mibitrans_model_instantreaction
    if isinstance(expected, float):
        assert model.sample(x, y, t) == pytest.approx(expected)
    elif expected is ValueError or expected is TypeError:
        with pytest.raises(expected):
            model.sample(x, y, t)



"""Author: Jorrit Bakker.

Module evaluating if a dictionary contains all required (correct) parameters for analysis
"""

import warnings
import numpy as np
import mibitrans


def _check_numeric(parameter: str, value):
    """Check if a variable is numerical and if it is positive."""
    if isinstance(value, (float, int, np.floating, np.integer)):
        return None
    else:
        return TypeError(f"{parameter} must be a float, but is {type(value)} instead.")


def _check_numeric_positive(parameter: str, value):
    """Check if a variable is numerical and if it is positive."""
    is_float = _check_numeric(parameter, value)
    if is_float is None:
        if value >= 0:
            return None
        else:
            return DomainValueError(f"{parameter} must be >= 0")
    else:
        return is_float


def _check_numeric_fraction(parameter: str, value):
    """Check if a variable is numerical and if it is between 0 and 1."""
    is_float = _check_numeric(parameter, value)
    if is_float is None:
        if 0 <= value <= 1:
            return None
        else:
            return DomainValueError(f"{parameter} must be between 0 and 1")
    else:
        return is_float


def _check_numeric_retardation(parameter: str, value):
    """Check if a variable is numerical and if it is 1 or larger."""
    is_float = _check_numeric(parameter, value)
    if is_float is None:
        if value >= 1:
            return None
        else:
            return DomainValueError(f"{parameter} must be 1 or larger.")
    else:
        return TypeError(f"{parameter} must be a float, but is {type(value)} instead.")


def _check_array_list_numeric_positive(parameter: str, value):
    """Check if variable is numpy array, list, or numerical, if it is positive and if an array is 1-dimensional."""
    if isinstance(value, np.ndarray):
        if len(value.shape) == 1:
            if all(value >= 0):
                return None
            else:
                return DomainValueError(f"All values in {parameter} should be >= 0.")
        else:
            return ValueError(
                f"{parameter} should be a float, list or a 1-dimensional array,"
                f"but input array is {len(value.shape)}-dimensional."
            )

    elif isinstance(value, list):
        if all(isinstance(element, (float, int, np.floating, np.integer)) for element in value):
            if all(element >= 0 for element in value):
                return None
            else:
                return DomainValueError(f"All values in {parameter} should be >= 0.")
        else:
            return TypeError(f"All elements of {parameter} should be a float.")

    elif isinstance(value, (float, int, np.floating)):
        if value >= 0:
            return None
        else:
            return DomainValueError(f"{parameter} must be >= 0")

    else:
        return TypeError(f"{parameter} must be a float, list or numpy array, but is {type(value)} instead.")


def validate_source_zones(boundary, concentration):
    """Validate and adapt input of source_zone_boundary and source_zone_concentration arrays."""
    # Ensure boundary and concentration are numpy arrays
    if isinstance(boundary, (float, int, np.floating, np.integer)):
        boundary = np.array([boundary])
    else:
        boundary = np.array(boundary)

    if isinstance(concentration, (float, int, np.floating, np.integer)):
        concentration = np.array([concentration], dtype=float)
    else:
        concentration = np.array(concentration, dtype=float)

    # Each given source zone boundary should have a given concentration, and vice versa
    if boundary.shape != concentration.shape:
        raise ValueError(
            f"Length of source zone boundary ({len(boundary)}) and source zone concentration "
            f"({len(concentration)}) do not match. Make sure they are of equal length."
        )

    # Reorder source zone locations if they are not given in order from close to far from source zone center
    if len(boundary) > 1:
        if not all(boundary[:-1] <= boundary[1:]):
            sort_location = np.argsort(boundary)
            boundary.sort()
            concentration = concentration[sort_location]
            warnings.warn(
                "Source zone boundary locations should be ordered by distance from source zone center. "
                "Zone boundaries and concentrations have consequently been reordered as follows:"
                f"Source zone boundaries: {boundary}"
                f"Source zone concentrations: {concentration}"
            )
        # Superposition method only works if the zone closer to the center has higher concentration than outer zones
        if not all(concentration[:-1] > concentration[1:]):
            raise ValueError(
                "Source zone concentrations should be in descending order; no source zone can have a concentration "
                "higher than the concentration of a zone closer to source center, due to the superposition method."
            )
    return boundary, concentration


def _check_total_mass(parameter: str, value):
    """Check variable properties of total source mass specifically."""
    if isinstance(value, str):
        if "inf" not in value:
            return ValueError(f"{value} is not understood. For infinite source mass, use 'infinite' or 'np.inf'.")
        else:
            return None
    elif _check_numeric(parameter, value) is None:
        if value >= 0:
            return None
        else:
            return DomainValueError(f"{parameter} must be >= 0, or set to 'infinite'.")
    else:
        return TypeError(f"{parameter} must be a float or 'infinite', but is {type(value)} instead.")


def check_dictionary(value):
    """Check if variable is a dictionary, and raise an error if it is not."""
    if not isinstance(value, dict):
        raise TypeError(f"Input must be a dict, but is {type(value)} instead.")


def _check_dataclass(parameter, value, expected_type):
    """Check if variable is of the given type, and raise an error if it is not."""
    if isinstance(value, expected_type):
        return None
    else:
        return TypeError(f"{parameter} must be of type {expected_type}, but is {type(value)} instead.")


def _check_electron_acceptor(value):
    """Check if variable is an ElectronAcceptors dataclass, list, array or dictionary, raise an error if it is not."""
    if isinstance(value, mibitrans.data.parameter_information.ElectronAcceptors):
        return None
    elif isinstance(value, (list, np.ndarray, dict)):
        if len(value) != 5:
            return ValueError(
                f"Input for electron_acceptors as list, array or dictionary must have an entry for each electron "
                f"acceptor, of which there are five utilized by this model. The current input has {len(value)} "
                f"entries instead."
            )
        else:
            return None

    else:
        return TypeError(
            f"electron_acceptors must be of type {mibitrans.data.parameter_information.ElectronAcceptors},"
            f" or alternatively as list, numpy array or dictionary containing electron acceptor "
            f"concentrations. But is {type(value)} instead."
        )


def check_model_type(parameter, allowed_model_types):
    """Check if variable is of the given allowed model types, and raise an error if it is not."""
    if not isinstance(parameter, allowed_model_types):
        if isinstance(allowed_model_types, tuple):
            raise TypeError(
                f"Input argument model should be subclass of {allowed_model_types}, but is {type(parameter)} instead."
            )
        else:
            raise TypeError(
                f"Input argument model should be in {allowed_model_types.__subclasses__()}, "
                f"but is {type(parameter)} instead."
            )


def check_time_in_domain(model, time):
    """Check if time input is valid, and returns the index of nearest time."""
    if time is not None:
        error = _check_numeric_positive("time", time)
        if error is not None:
            raise error
        elif time > np.max(model.t):
            warnings.warn(
                f"Desired time is larger than maximum time of model ({time} > {np.max(model.t)}). Using maximum time "
                f"of model instead."
            )
            time_pos = len(model.t) - 1
        else:
            time_pos = np.argmin(abs(model.t - time))
    else:
        time_pos = len(model.t) - 1
    return time_pos


def check_y_in_domain(model, y_position):
    """Check if y-position input is valid, and returns the index of nearest y position."""
    error = _check_numeric("y_position", y_position)
    if error is not None:
        raise error
    if y_position > np.max(model.y):
        warnings.warn(
            f"Desired y position is outside of model domain (abs({y_position}) > {np.max(model.y)}). "
            f"Using closest position inside model domain instead."
        )

    y_pos = np.argmin(abs(model.y - y_position))
    return y_pos


def check_x_in_domain(model, x_position):
    """Check if x-position input is valid, and returns the index of nearest x position."""
    error = _check_numeric_positive("x_position", x_position)
    if error is not None:
        raise error
    if x_position > np.max(model.x):
        warnings.warn(
            f"Desired x position is outside of model domain ({x_position} > {np.max(model.x)}). "
            f"Using closest position inside model domain instead."
        )

    x_pos = np.argmin(abs(model.x - x_position))
    return x_pos


def validate_input_values(parameter, value):
    """Validate if input parameter is of correct type and in correct domain."""
    match parameter:
        # Any input for verbose argument is fine; if set to anything other than False, 0 or None, verbose is on
        case "verbose":
            error = None
        case "_on_change":
            error = None
        # Specific check for retardation, which has domain >= 1
        case "retardation":
            error = _check_numeric_retardation(parameter, value)
        # Specific check for total mass, which can be a positive float, or a specific string
        case "total_mass":
            error = _check_total_mass(parameter, value)
        # Specific check for electron acceptor utilization factor, which should be UtilizationFactor dataclass
        case "utilization_factor":
            error = _check_dataclass(parameter, value, mibitrans.data.parameter_information.UtilizationFactor)
        # Parameters which can be any float value
        case "y_position":
            error = _check_numeric(parameter, value)
        # Parameters which have domain [0,1]
        case "porosity" | "fraction_organic_carbon":
            error = _check_numeric_fraction(parameter, value)
        # Parameters which are input as single values, lists or numpy arrays
        case "source_zone_boundary" | "source_zone_concentration":
            error = _check_array_list_numeric_positive(parameter, value)
        case "electron_acceptors":
            error = _check_electron_acceptor(value)
        # All other parameters are checked as floats on positive domain
        case _:
            error = _check_numeric_positive(parameter, value)

    if error and (value is not None):
        raise error


class DomainValueError(Exception):
    """Exception raised for values that are outside their possible domain.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        """Initialize error class."""
        self.message = message
        super().__init__(self.message)


class MissingValueError(Exception):
    """Exception raised when one or more required parameters are missing.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        """Initialize error class."""
        self.message = message
        super().__init__(self.message)



"""Author: Jorrit Bakker.

Module containing various methods that takes a dictionary of parameters as input and calculates the proper values that
can be used in transport equations.
"""

import numpy as np
from mibitrans.data.parameter_information import util_to_conc_name


def calculate_utilization(model):
    """Function that calculates relative use of electron acceptors in biodegradation of BTEX."""
    util_factor = model._utilization_factor.dictionary
    biodeg_array = np.zeros(len(list(util_factor.keys())))
    util_array = np.zeros(len(biodeg_array))

    for i, (key, value) in enumerate(util_factor.items()):
        biodeg_array[i] = getattr(model._electron_acceptors, util_to_conc_name[key]) / value
        util_array[i] = value

    biodegradation_capacity = np.sum(biodeg_array)
    fraction_total = biodeg_array / biodegradation_capacity
    mass_fraction = fraction_total * util_array

    return mass_fraction


def calculate_discharge_and_average_source_zone_concentration(model):
    """Calculate discharge through source zone and average source zone concentration, returned in respective order."""
    if model.mode == "instant_reaction":
        bc = model.biodegradation_capacity
    else:
        bc = 0
    y_src = np.zeros(len(model.source_parameters.source_zone_boundary) + 1)
    y_src[1:] = model.source_parameters.source_zone_boundary
    c_src = model.source_parameters.source_zone_concentration
    Q = (
        model.hydrological_parameters.velocity
        * model.hydrological_parameters.porosity
        * model.source_parameters.depth
        * np.max(y_src)
        * 2
    )

    weighted_conc = np.zeros(len(model.source_parameters.source_zone_boundary))
    for i in range(len(model.source_parameters.source_zone_boundary)):
        weighted_conc[i] = (y_src[i + 1] - y_src[i]) * c_src[i]

    c0_avg = bc + np.sum(weighted_conc) / np.max(y_src)

    return Q, c0_avg
